from .blackbody import (
    Blackbody,
    ModifiedBlackbody,
    NonDissipativePhotosphere,
    NonDissipativePhotosphere_Deep,
)
from .functions import (
    DiracDelta,
    Exponential_cutoff,
    Log_parabola,
    Sin,
    StepFunction,
    StepFunctionUpper,
    has_gsl,
    has_naima,
    GenericFunction
)

if has_naima:
    from .functions import Synchrotron

if has_gsl:

    from .functions import Cutoff_powerlaw_flux

from .absorption import PhAbs, TbAbs, WAbs, has_ebltable
from .apec import has_atomdb

if has_ebltable:
    from .absorption import EBLattenuation


from .extinction import Standard_Rv, ZDust
from .polynomials import (
    Constant,
    Cubic,
    Line,
    Quadratic,
    Quartic,
    get_polynomial,
)
from .powerlaws import (
    Band,
    Band_Calderone,
    Band_grbm,
    Broken_powerlaw,
    Cutoff_powerlaw,
    Cutoff_powerlaw_Ep,
    Inverse_cutoff_powerlaw,
    Powerlaw,
    Powerlaw_Eflux,
    Powerlaw_flux,
    SmoothlyBrokenPowerLaw,
    Super_cutoff_powerlaw,
    DoubleSmoothlyBrokenPowerlaw,
)

if has_atomdb:

    from .apec import APEC, VAPEC


__all__ = [
    "Band",
    "Band_Calderone",
    "Band_grbm",
    "Broken_powerlaw",
    "Cutoff_powerlaw",
    "Cutoff_powerlaw_Ep",
    "Inverse_cutoff_powerlaw",
    "Powerlaw",
    "Powerlaw_Eflux",
    "Powerlaw_flux",
    "SmoothlyBrokenPowerLaw",
    "DoubleSmoothlyBrokenPowerlaw",
    "Super_cutoff_powerlaw",
    "Constant",
    "Cubic",
    "DiracDelta",
    "Exponential_cutoff",
    "Line",
    "Quadratic",
    "Sin",
    "StepFunction",
    "StepFunctionUpper",
    "GenericFunction",
    "PhAbs",
    "TbAbs",
    "WAbs",
    "Log_parabola",
    "Blackbody",
    "ModifiedBlackbody",
    "NonDissipativePhotosphere",
    "NonDissipativePhotosphere_Deep",
    "Quartic",
    "get_polynomial",
    "ZDust",
    "Standard_Rv",
]

if has_atomdb:
    __all__.extend(["APEC", "VAPEC"])


if has_gsl:

    __all__.extend(["Cutoff_powerlaw_flux"])


if has_naima:

    __all__.extend(["Synchrotron"])


if has_ebltable:

    __all__.extend(["EBLattenuation"])



__author__ = "giacomov"

import collections

from astropy import coordinates

from astromodels.core.parameter import Parameter
from astromodels.core.tree import Node


class WrongCoordinatePair(ValueError):
    pass


class IllegalCoordinateValue(ValueError):
    pass


class WrongCoordinateSystem(ValueError):
    pass


class SkyDirection(Node):
    """
    This is essentially a wrapper around astropy.coordinates.SkyCoord with a possibility for
    being serialized and deserialized with YAML.
    """

    def __init__(self, ra=None, dec=None, l=None, b=None, equinox="J2000"):
        """

        :param ra: Right Ascension in degrees
        :param dec: Declination in degrees
        :param l: Galactic latitude in degrees
        :param b: Galactic longitude in degrees
        :param equinox: string
        :return:
        """

        self._equinox = equinox

        # Create the node

        Node.__init__(self, "position")

        # Check that we have the right pairs of coordinates

        if ra is not None and dec is not None:

            # This goes against duck typing, but it is needed to provide a means of initiating this class
            # with either Parameter instances or just floats

            # Try to transform it to float, if it works than we transform it to a parameter

            ra = self._get_parameter_from_input(ra, 0, 360, "ra", "Right Ascension")

            dec = self._get_parameter_from_input(dec, -90, 90, "dec", "Declination")

            self._coord_type = "equatorial"

            self._add_child(ra)
            self._add_child(dec)

        elif l is not None and b is not None:

            # This goes against duck typing, but it is needed to provide a means of initiating this class
            # with either Parameter instances or just floats

            # Try to transform it to float, if it works than we transform it to a parameter

            l = self._get_parameter_from_input(l, 0, 360, "l", "Galactic longitude")

            b = self._get_parameter_from_input(b, -90, 90, "b", "Galactic latitude")

            self._coord_type = "galactic"
            self._add_child(l)
            self._add_child(b)

        else:

            raise WrongCoordinatePair("You have to specify either (ra, dec) or (l, b).")

    @staticmethod
    def _get_parameter_from_input(number_or_parameter, minimum, maximum, what, desc):

        # Try to transform it to float, if it works than we transform it to a parameter

        try:

            number_or_parameter = float(number_or_parameter)

        except TypeError:

            assert isinstance(number_or_parameter, Parameter), (
                "%s must be either a number or a " "parameter instance" % what
            )

            # So this is a Parameter instance already. Enforce that it has the right maximum and minimum

            parameter = number_or_parameter

            assert (
                parameter.min_value >= minimum
            ), "%s must have a minimum greater than or equal to %s" % (
                what,
                minimum,
            )
            assert (
                parameter.max_value <= maximum
            ), "%s must have a maximum less than or equal to %s" % (
                what,
                maximum,
            )

        else:

            # This was a float. Enforce that it has a legal value

            assert (
                minimum <= number_or_parameter <= maximum
            ), "%s cannot have a value of %s, " "it must be %s <= %s <= %s" % (
                what,
                number_or_parameter,
                minimum,
                what,
                maximum,
            )

            parameter = Parameter(
                what,
                number_or_parameter,
                desc=desc,
                min_value=minimum,
                max_value=maximum,
                unit="deg",
                free=False,
            )

        return parameter

    def get_ra(self):
        """
        Get R.A. corresponding to the current position (ICRS, J2000)

        :return: Right Ascension
        """

        try:

            return self.ra.value

        except AttributeError:

            # Transform from L,B to R.A., Dec

            return self.sky_coord.transform_to("icrs").ra.value

    def get_dec(self):
        """
        Get Dec. corresponding to the current position (ICRS, J2000)

        :return: Declination
        """

        try:

            return self.dec.value

        except AttributeError:

            # Transform from L,B to R.A., Dec

            return self.sky_coord.transform_to("icrs").dec.value

    def get_l(self):
        """
        Get Galactic Longitude (l) corresponding to the current position

        :return: Galactic Longitude
        """

        try:

            return self.l.value

        except AttributeError:

            # Transform from L,B to R.A., Dec

            return self.sky_coord.transform_to("galactic").l.value

    def get_b(self):
        """
        Get Galactic latitude (b) corresponding to the current position

        :return: Latitude
        """

        try:

            return self.b.value

        except AttributeError:

            # Transform from L,B to R.A., Dec

            return self.sky_coord.transform_to("galactic").b.value

    def _get_sky_coord(self):

        if self._coord_type == "galactic":

            l = self.l.value
            b = self.b.value

            return coordinates.SkyCoord(
                l=l, b=b, frame="galactic", equinox=self._equinox, unit="deg"
            )

        else:

            ra = self.ra.value
            dec = self.dec.value

            return coordinates.SkyCoord(
                ra=ra, dec=dec, frame="icrs", equinox=self._equinox, unit="deg"
            )

    @property
    def sky_coord(self):
        """
        Return an instance of astropy.coordinates.SkyCoord which can be used to make all transformations supported
        by it

        :return: astropy.coordinates.SkyCoord
        """
        return self._get_sky_coord()

    @property
    def parameters(self):
        """
        Get the dictionary of parameters (either ra,dec or l,b)

        :return: dictionary of parameters
        """

        if self._coord_type == "galactic":

            return collections.OrderedDict((("l", self.l), ("b", self.b)))

        else:

            return collections.OrderedDict((("ra", self.ra), ("dec", self.dec)))

    @property
    def equinox(self):
        """
        Returns the equinox for the coordinates.

        :return:
        """
        return self._equinox

    def to_dict(self, minimal=False):

        data = collections.OrderedDict()

        if self._coord_type == "equatorial":

            data["ra"] = self.ra.to_dict(minimal)
            data["dec"] = self.dec.to_dict(minimal)
            data["equinox"] = self._equinox

        else:

            data["l"] = self.l.to_dict(minimal)
            data["b"] = self.b.to_dict(minimal)
            data["equinox"] = self._equinox

        return data

    def fix(self):
        """
        Fix the parameters with the coordinates (either ra,dec or l,b depending on how the class
        has been instanced)

        """

        if self._coord_type == "equatorial":

            self.ra.fix = True
            self.dec.fix = True

        else:

            self.l.fix = True
            self.b.fix = True

    def free(self):
        """
        Free the parameters with the coordinates (either ra,dec or l,b depending on how the class
        has been instanced)

        """

        if self._coord_type == "equatorial":

            self.ra.fix = False
            self.dec.fix = False

        else:

            self.l.fix = False
            self.b.fix = False

    @classmethod
    def from_dict(cls, data):

        return cls(**data)

    def _repr__base(self, rich_output):

        if self._coord_type == "equatorial":

            representation = "Sky direction (R.A., Dec.) = (%.5f, %.5f) (%s)" % (
                self.ra.value,
                self.dec.value,
                self.equinox,
            )

        else:

            representation = "Sky direction (l, b) = (%.5f, %.5f) (%s)" % (
                self.l.value,
                self.b.value,
                self.equinox,
            )

        return representation



from __future__ import absolute_import

import os

from ._version import get_versions

# Import the version

#
#

if os.environ.get("ASTROMODELS_DEBUG", None) is None:

    from .utils.configuration import astromodels_config, show_configuration
    from .core.memoization import use_astromodels_memoization
    from .core.model import Model
    from .core.model_parser import clone_model, load_model
    from .core.parameter import (
        IndependentVariable,
        Parameter,
        SettingOutOfBounds,
        turn_off_parameter_transforms,
    )
    from .core.polarization import LinearPolarization, StokesPolarization
    from .core.serialization import *
    from .core.spectral_component import SpectralComponent
    from .core.units import get_units
    from .functions import (
        Asymm_Gaussian_on_sphere,
        Band,
        Band_Calderone,
        Band_grbm,
        Blackbody,
        Broken_powerlaw,
        Cauchy,
        Constant,
        Continuous_injection_diffusion,
        Continuous_injection_diffusion_ellipse,
        Continuous_injection_diffusion_legacy,
        Cosine_Prior,
        Cubic,
        Cutoff_powerlaw,
        Cutoff_powerlaw_Ep,
        DiracDelta,
        Disk_on_sphere,
        DMFitFunction,
        DMSpectra,
        Ellipse_on_sphere,
        Exponential_cutoff,
        Function1D,
        Function2D,
        Function3D,
        FunctionMeta,
        GalPropTemplate_3D,
        Gaussian,
        Gaussian_on_sphere,
        Inverse_cutoff_powerlaw,
        Latitude_galactic_diffuse,
        Line,
        Log_normal,
        Log_parabola,
        Log_uniform_prior,
        MissingDataFile,
        ModelAssertionViolation,
        ModifiedBlackbody,
        NonDissipativePhotosphere,
        NonDissipativePhotosphere_Deep,
        PhAbs,
        Power_law_on_sphere,
        Powerlaw,
        Powerlaw_Eflux,
        Powerlaw_flux,
        Quadratic,
        Quartic,
        Sin,
        SmoothlyBrokenPowerLaw,
        DoubleSmoothlyBrokenPowerlaw,
        SpatialTemplate_2D,
        Standard_Rv,
        StepFunction,
        StepFunctionUpper,
        GenericFunction,
        Super_cutoff_powerlaw,
        TbAbs,
        TemplateModel,
        TemplateModelFactory,
        Truncated_gaussian,
        Uniform_prior,
        WAbs,
        XSPECTableModel,
        ZDust,
        get_polynomial,
        has_ebltable,
        has_gsl,
        has_naima,
        has_atomdb,
    )

    if has_ebltable:

        from .functions import EBLattenuation

    if has_gsl:

        from .functions import Cutoff_powerlaw_flux

    if has_naima:

        from .functions import Synchrotron

    if has_atomdb:

        from .functions import APEC, VAPEC

    from .functions.function import get_function_class, list_functions
    from .sources import ExtendedSource, ParticleSource, PointSource

    astromodels_units = get_units()
    from astromodels.utils.logging import (
        setup_logger,
        update_logging_level,
        silence_warnings,
        activate_warnings,
    )

import astropy.units as u

log = setup_logger(__name__)

__version__ = get_versions()["version"]
del get_versions



import logging
from dataclasses import dataclass, field
from enum import IntEnum, Enum


# logging
class LoggingLevel(IntEnum):
    DEBUG = logging.DEBUG
    INFO = logging.INFO
    WARNING = logging.WARNING
    ERROR = logging.ERROR
    CRITICAL = logging.CRITICAL


@dataclass
class Logging:
    path: str = "~/.astromodels/log"
    developer: bool = "off"
    usr: bool = "on"
    console: bool = "on"
    level: LoggingLevel = LoggingLevel.INFO
    startup_warnings: bool = "on"
    info_style: str = "medium_spring_green"
    warn_style: str = "medium_orchid"
    error_style: str = "blink bold bright_red"
    debug_style: str = "blue_violet"
    message_style: str = "bold grey78"


class AbsTables(Enum):
    WILM = "WILM"
    ASPL = "ASPL"
    AG89 = "AG89"


class EBLTable(Enum):
    franceschini = "franceschini"
    kneiske = "kneiske"
    dominguez = "dominguez"
    inuoe = "inuoe"
    gilmore = "gilmore"


@dataclass
class AbsorptionModels:
    tbabs_table: AbsTables = AbsTables.WILM
    phabs_table: AbsTables = AbsTables.AG89
    ebl_table: EBLTable = EBLTable.dominguez


@dataclass
class Modeling:
    use_memoization: bool = True
    use_parameter_transforms: bool = True
    ignore_parameter_bounds: bool = False


@dataclass
class Config:
    logging: Logging = field(default_factory=Logging)
    absorption_models: AbsorptionModels = field(
        default_factory=AbsorptionModels
    )
    modeling: Modeling = field(default_factory=Modeling)



from __future__ import print_function

import gc
import os
from builtins import range

import astropy.units as u
import pytest

from astromodels.core.tree import Node

# os.environ["ASTROMODELS_DEBUG"] = "debug"

# from astromodels.core import node_ctype


def clean():

    gc.collect()


class _SimpleInheritance(Node):
    def __init__(self, name):
        print("INIT of _SimpleInheritance")

        self._placeholder = 2

        super(_SimpleInheritance, self).__init__(name)

    @property
    def placeholder(self):
        return self._placeholder


class _ComplexInheritance(Node):
    def __init__(self, name, min_value, max_value):

        super(_ComplexInheritance, self).__init__(name)

        self._min_value = min_value * u.Unit("1 / (keV * cm**2 * s)")
        self._max_value = max_value * u.Unit("1 / (keV * cm**2 * s)")

    @property
    def min_value(self):

        return self._min_value

    @property
    def max_value(self):

        return self._max_value


def test_constructor():

    n = Node(name="node1")

    assert n.name == "node1"

    with pytest.raises(TypeError):

        n2 = Node()

    clean()


def test_inheritance():
    t = _SimpleInheritance("test1")

    assert t.name == "test1"

    clean()


def test_add_child():

    t = _SimpleInheritance("test_add_child")

    with pytest.raises(TypeError):
        t._add_child("clara")

    clara = Node("clara")

    t._add_child(clara)

    t.plot_tree()

    assert t.clara == clara

    with pytest.raises(AttributeError):
        t.clara = "argh"

    assert t.clara == clara

    clean()


def test_add_children():
    t = _SimpleInheritance("children")

    t._add_children([Node("node1"), Node("node2")])

    clean()


def test_get_child():
    t = _SimpleInheritance("test")

    n = Node("node")

    t._add_child(n)

    assert t._get_child("node") == n

    with pytest.raises(KeyError):
        t._get_child("not existing")

    clean()


def test_hashing():

    node1 = Node("node1")
    node2 = Node("node2")
    node22 = Node("node22")
    node3 = Node("node3")

    d = {}

    # nodes as hashes

    d[node1] = 1
    d[node2] = 2

    d[node1] == 1
    d[node2] == 2

    # nodes in dicts

    d = {}

    d["node1"] = node1

    node1._add_child(node2)
    node1._add_child(node22)
    node2._add_child(node3)

    d = {}

    d["node1"] = node1
    d["node2"] = node2

    d = {}

    d = {}

    d[node1] = 1
    d[node1.node2] = 2

    d[node1] == 1
    d[node2] == 2


def test_get_children():
    node1 = Node("node1")
    node2 = Node("node2")
    node22 = Node("node22")
    node3 = Node("node3")

    node1._add_child(node2)
    node1._add_child(node22)
    node2._add_child(node3)

    assert node1._get_children() == (node2, node22)
    assert node2._get_children() == (node3,)

    clean()


def test_remove_child():
    t = _SimpleInheritance("test")

    n = Node("node")

    for i in range(1000):

        t._add_child(n)

        assert t._get_child("node") == n

        t._remove_child("node")

    with pytest.raises(AttributeError):
        print(t.node)

    with pytest.raises(KeyError):
        t._get_child("node")

    clean()

    t = _SimpleInheritance("test")

    n = Node("node")

    for i in range(1000):

        t._add_child(n)

        assert t._get_child("node") == n

        old = t._remove_child("node", delete=False)

        assert old == n

        assert old.is_root

    with pytest.raises(AttributeError):
        print(t.node)

    with pytest.raises(KeyError):
        t._get_child("node")

    clean()


def test_get_path():
    # Make a small tree

    node1 = Node("node1")
    node2 = Node("node2")
    node3 = Node("node3")

    node1._add_child(node2)
    node2._add_child(node3)

    print(node3.path)

    assert node3.path == "node1.node2.node3"

    clean()


def test_get_child_from_path():
    # Make a small tree

    node1 = Node("node1")
    node2 = Node("node2")
    node3 = Node("node3")

    node1._add_child(node2)
    node2._add_child(node3)

    assert node1._get_child_from_path("node2.node3") == node3

    clean()


def test_change_name():
    t = _SimpleInheritance("name1")

    assert t.name == "name1"

    with pytest.raises(AttributeError):
        t.name = "name2"

    assert t.name == "name1"

    t._change_name("name2")

    assert t.name == "name2"

    clean()


def test_pickle():

    print("\n\n\n\n")

    import pickle

    root = Node("root")
    node = Node("node")

    root._add_child(node)

    d = pickle.dumps(root)

    root2 = pickle.loads(d)

    assert root2.node.path == "root.node"
    assert root2.node.name == "node"

    # Now test pickling a subclass of Node

    root = _SimpleInheritance("root")

    root._placeholder = 5.3

    node = _SimpleInheritance("node")

    root._add_child(node)

    d = pickle.dumps(root)

    root2 = pickle.loads(d)

    assert root2.node.path == "root.node"
    assert root2.node.name == "node"

    print(root.placeholder)

    assert root2.placeholder == root.placeholder

    rootc = _ComplexInheritance("root", -1.0, 1.0)
    nodec = _ComplexInheritance("node", -5.0, 5.0)

    rootc._add_child(nodec)

    d = pickle.dumps(rootc)

    root2c = pickle.loads(d)

    print(root2c.min_value)

    clean()



import collections
import os
import uuid
from typing import Any, Dict, List, Optional, Union

from future import standard_library

from astromodels.core.node_type import NodeBase
from astromodels.utils.io import display
from astromodels.utils.logging import setup_logger
from astromodels.utils.valid_variable import is_valid_variable_name

standard_library.install_aliases()


class DuplicatedNode(Exception):
    pass


class ProtectedAttribute(RuntimeError):
    pass


class NonExistingAttribute(RuntimeWarning):
    pass


log = setup_logger(__name__)


class Node(NodeBase):

    # This apparently dumb constructor is needed otherwise pickle will fail

    def __init__(self, name: str):

        if not is_valid_variable_name(name):

            log.error(
                "Illegal characters in name %s. You can only use letters and numbers, "
                "and _" % name
            )

            raise AssertionError()

        if not name != "name":

            log.error("You cannot call a node 'name', it is reserved.")

            raise AssertionError()

        NodeBase.__init__(self, name)

        #########################################################################

    def __hash__(self):

        # return the hash of the path
        # anytime we need a hash should
        # be after a model is finalized
        # so it should be static

        return hash(self._get_path())

    def to_dict(self, minimal: bool = False) -> Dict[str, Any]:
        """ """
        this_dict: Dict[str, Any] = collections.OrderedDict()

        for child in self._get_children():

            this_dict[child.name] = child.to_dict(minimal)

        return this_dict

    # This is used by dir() and by the autocompletion in Ipython

    def __dir__(self):

        # Get the names of the attributes of the class
        l = list(self.__class__.__dict__.keys())

        # Get all the children
        l.extend([child.name for child in self._get_children()])

        return l

    def _repr__base(self, rich_output):

        return f"Node with name {self.name}"

    def __repr__(self):
        """
        Textual representation for console

        :return: representation
        """

        return self._repr__base(rich_output=False)

    def _repr_html_(self):
        """
        HTML representation for the IPython notebook

        :return: HTML representation
        """

        return self._repr__base(rich_output=True)

    def display(self):
        """
        Display information about the point source.

        :return: (none)
        """

        # This will automatically choose the best representation among repr and repr_html

        display(self)



from __future__ import division, print_function

import astropy.units as u
import numpy as np
import numpy.testing as npt
import pytest

from astromodels.core.spectral_component import SpectralComponent
from astromodels.functions import (
    Band,
    Blackbody,
    Exponential_cutoff,
    Log_parabola,
    Powerlaw,
)
from astromodels.functions.functions_1D.functions import _ComplexTestFunction, GenericFunction

try:
    from astromodels.functions import PhAbs, TbAbs, WAbs

    has_abs_models = True

except:

    has_abs_models = False


from astromodels.core.model import Model
from astromodels.core.model_parser import clone_model, load_model
from astromodels.sources.particle_source import ParticleSource
from astromodels.sources.point_source import PointSource

try:

    from astromodels.xspec import XS_phabs, XS_powerlaw

except:

    has_xspec = False

else:

    has_xspec = True

try:

    from astromodels.functions import EBLattenuation

except:

    has_ebl = False

else:

    has_ebl = True

from astromodels.functions.function import _known_functions
from astromodels.functions.priors import *

__author__ = "giacomov"


_multiplicative_models = ["PhAbs", "TbAbs", "WAbs", "EBLattenuation", "ZDust"]


def test_constructor():

    # RA, Dec and L,B of the same point in the sky

    ra, dec = (125.6, -75.3)
    l, b = (288.44190139183564, -20.717313145391525)

    # This should throw as we are using Powerlaw instead of Powerlaw()
    with pytest.raises(TypeError):

        _ = PointSource("my_source", ra, dec, Powerlaw)

    # Init with RA, Dec

    point_source1 = PointSource("my_source", ra, dec, Powerlaw())

    assert point_source1.position.get_ra() == ra
    assert point_source1.position.get_dec() == dec

    assert abs(point_source1.position.get_l() - l) < 1e-7
    assert abs(point_source1.position.get_b() - b) < 1e-7

    assert point_source1.position.ra.value == ra
    assert point_source1.position.dec.value == dec

    # Verify that the position is fixed by default
    assert point_source1.position.ra.fix
    assert point_source1.position.dec.fix

    # Init with l,b

    point_source2 = PointSource("my_source", l=l, b=b, spectral_shape=Powerlaw())

    assert point_source2.position.get_l() == l
    assert point_source2.position.get_b() == b

    assert abs(point_source2.position.get_ra() - ra) < 1e-7
    assert abs(point_source2.position.get_dec() - dec) < 1e-7

    assert point_source2.position.l.value == l
    assert point_source2.position.b.value == b

    # Verify that the position is fixed by default
    assert point_source2.position.l.fix
    assert point_source2.position.b.fix

    # Multi-component

    po1 = Powerlaw()
    po2 = Powerlaw()

    c1 = SpectralComponent("component1", po1)
    c2 = SpectralComponent("component2", po2)

    point_source3 = PointSource("test_source", ra, dec, components=[c1, c2])

    assert np.all(point_source3.spectrum.component1([1, 2, 3]) == po1([1, 2, 3]))
    assert np.all(point_source3.spectrum.component2([1, 2, 3]) == po2([1, 2, 3]))

    with pytest.raises(AssertionError):

        # Illegal RA

        _ = PointSource("test", 720.0, -15.0, components=[c1, c2])

    with pytest.raises(AssertionError):
        # Illegal Dec

        _ = PointSource("test", 120.0, 180.0, components=[c1, c2])

    with pytest.raises(AssertionError):
        # Illegal l

        _ = PointSource("test", l=-195, b=-15.0, components=[c1, c2])

    with pytest.raises(AssertionError):
        # Illegal b

        _ = PointSource("test", l=120.0, b=-180.0, components=[c1, c2])


def test_call():

    # Multi-component

    po1 = Powerlaw()
    po2 = Powerlaw()

    c1 = SpectralComponent("component1", po1)
    c2 = SpectralComponent("component2", po2)

    point_source = PointSource("test_source", 125.4, -22.3, components=[c1, c2])

    assert np.all(point_source.spectrum.component1([1, 2, 3]) == po1([1, 2, 3]))
    assert np.all(point_source.spectrum.component2([1, 2, 3]) == po2([1, 2, 3]))

    one = point_source.spectrum.component1([1, 2, 3])
    two = point_source.spectrum.component2([1, 2, 3])

    assert np.all(np.abs(one + two - point_source([1, 2, 3])) == 0)


def test_call_with_units():

    po = Powerlaw()

    result = po(1.0)

    assert result.ndim == 0

    with pytest.raises(AssertionError):

        # This raises because the units of the function have not been set up

        _ = po(1.0 * u.keV)

    # Use the function as a spectrum
    ps = PointSource("test", 0, 0, po)

    result = po(1.0 * u.keV)

    assert isinstance(result, u.Quantity)

    result = po(np.array([1, 2, 3]) * u.keV)

    assert isinstance(result, u.Quantity)

    # Now test all the functions
    def test_one(class_type):

        if class_type == _ComplexTestFunction:

            instance = class_type(file_name="test.txt")

        else:

            instance = class_type()

        if not instance.is_prior and not instance.name == "GenericFunction":

            # if we have fixed x_units then we will use those
            # in the test

            if instance.has_fixed_units():

                x_unit_to_use = instance.fixed_units[0]

            else:

                x_unit_to_use = u.keV

            # Use the function as a spectrum
            ps = PointSource("test", 0, 0, instance)

            if instance.name in ["Synchrotron", "_ComplexTestFunction"]:

                # we should not do it this way

                p = Powerlaw()

                particleSource = ParticleSource("particles", p)

                instance.set_particle_distribution(p)

            # elif instance.name in ["PhAbs", "TbAbs"]:

            #     instance

            result = ps(1.0)

            assert isinstance(result, float)

            result = ps(1.0 * x_unit_to_use)

            assert isinstance(result, u.Quantity)

            result = ps(np.array([1, 2, 3]) * x_unit_to_use)

            assert isinstance(result, u.Quantity)

            if instance.name in ["Synchrotron", "_ComplexTestFunction"]:
                model = Model(particleSource, ps)

            else:
                model = Model(ps)

            new_model = clone_model(model)

            new_result = new_model["test"](np.array([1, 2, 3]) * x_unit_to_use)

            assert np.all(new_result == result)

            model.save("__test.yml", overwrite=True)

            new_model = load_model("__test.yml")

            new_result = new_model["test"](np.array([1, 2, 3]) * x_unit_to_use)

            assert np.all(new_result == result)

        else:

            print("Skipping prior function")

    for key in _known_functions:

        this_function = _known_functions[key]

        # Test only the power law of XSpec, which is the only one we know we can test at 1 keV

        if (
            key.find("XS") == 0
            and key != "XS_powerlaw"
            or (key in _multiplicative_models)
        ):

            # An XSpec model. Test it only if it's a power law (the others might need other parameters during
            # initialization)

            continue

        if key.find("TemplateModel") == 0:

            # The TemplateModel function has its own test

            continue

        if key.find("Synchrotron") == 0:

            # Naima Synchtron function should have its own test

            continue

        if this_function._n_dim == 1:

            print("testing %s ..." % key)

            test_one(_known_functions[key])


def test_call_with_composite_function_with_units():
    def one_test(spectrum):

        print("Testing %s" % spectrum.expression)

        # # if we have fixed x_units then we will use those
        # # in the test
        #
        # if spectrum.expression.has_fixed_units():
        #
        #     x_unit_to_use, y_unit_to_use = spectrum.expression.fixed_units[0]
        #
        # else:

        x_unit_to_use = u.keV

        pts = PointSource("test", ra=0, dec=0, spectral_shape=spectrum)

        res = pts(np.array([100.0, 200.0]) * x_unit_to_use)

        # This will fail if the units are wrong
        res.to(old_div(1, (u.keV * u.cm**2 * u.s)))

    # Test a simple composition

    spectrum = Powerlaw() * Exponential_cutoff()

    one_test(spectrum)

    spectrum = Band() + Blackbody()

    one_test(spectrum)

    # Test a more complicate composition

    spectrum = Powerlaw() * Exponential_cutoff() + Blackbody()

    one_test(spectrum)

    spectrum = Powerlaw() * Exponential_cutoff() * Exponential_cutoff() + Blackbody()

    one_test(spectrum)

    # test the absorption models

    if has_abs_models:

        spectrum = PhAbs() * Powerlaw()

        one_test(spectrum)

        spectrum = TbAbs() * Powerlaw()

        one_test(spectrum)

        spectrum = WAbs() * Powerlaw()

        one_test(spectrum)

    if has_xspec:

        spectrum = XS_phabs() * Powerlaw()

        one_test(spectrum)

        spectrum = XS_phabs() * XS_powerlaw()

        one_test(spectrum)

        spectrum = XS_phabs() * XS_powerlaw() * XS_phabs()

        one_test(spectrum)

        spectrum = XS_phabs() * XS_powerlaw() * XS_phabs() + Blackbody()

        one_test(spectrum)

        spectrum = XS_phabs() * XS_powerlaw() * XS_phabs() + XS_powerlaw()

        one_test(spectrum)

    if has_ebl:

        spectrum = Powerlaw() * EBLattenuation()

        one_test(spectrum)


def test_free_param():

    spectrum = Log_parabola()
    source = PointSource("test_source", ra=123.4, dec=56.7, spectral_shape=spectrum)

    parameters = [
        spectrum.alpha,
        spectrum.beta,
        spectrum.piv,
        spectrum.K,
        source.position.ra,
        source.position.dec,
    ]

    for param in parameters:
        param.free = False

    assert not source.has_free_parameters

    assert len(source.free_parameters) == 0

    assert len(source.parameters) > len(source.free_parameters)

    for i, param in enumerate(parameters):
        param.free = True
        assert len(source.free_parameters) == i + 1

    assert source.has_free_parameters


def test_local_deriv():

    p = Powerlaw(index=-2.0)

    npt.assert_allclose(-2.0, p.local_spectral_index(np.logspace(1, 3, 10)))



from __future__ import division

import hashlib

import astropy.units as u
import numpy as np
from astropy.coordinates import ICRS, BaseCoordinateFrame, SkyCoord
from astropy.io import fits
from past.utils import old_div
from scipy.interpolate import RegularGridInterpolator

from astromodels.functions.function import Function3D, FunctionMeta
from astromodels.utils.angular_distance import angular_distance_fast


class Continuous_injection_diffusion_ellipse(Function3D, metaclass=FunctionMeta):
    r"""
    description :

        Positron and electrons diffusing away from the accelerator

    latex : $\left(\frac{180^\circ}{\pi}\right)^2 \frac{1.2154}{\sqrt{\pi^3} r_{\rm diff} ({\rm angsep} ({\rm x, y, lon_0, lat_0})+0.06 r_{\rm diff} )} \, {\rm exp}\left(-\frac{{\rm angsep}^2 ({\rm x, y, lon_0, lat_0})}{r_{\rm diff} ^2} \right)$

    parameters :

        lon0 :

            desc : Longitude of the center of the source
            initial value : 0.0
            min : 0.0
            max : 360.0

        lat0 :

            desc : Latitude of the center of the source
            initial value : 0.0
            min : -90.0
            max : 90.0

        rdiff0 :

            desc : Projected diffusion radius. The maximum allowed value is used to define the truncation radius.
            initial value : 1.0
            min : 0
            max : 20

        delta :

            desc : index for the diffusion coefficient
            initial value : 0.5
            min : 0.3
            max : 0.6
            fix : yes

        b :

            desc : b field strength in uG
            initial value : 3
            min : 1
            max : 10.
            fix : yes

        piv :

            desc : Pivot for the diffusion radius
            initial value : 2e10
            min : 0
            fix : yes

        piv2 :

            desc : Pivot for converting gamma energy to electron energy (always be 1 TeV)
            initial value : 1e9
            min : 0
            fix : yes

        incl :

            desc : inclination of semimajoraxis to a line of constant latitude
            initial value : 0.0
            min : -90.0
            max : 90.0
            fix : yes

        elongation :

            desc : elongation of the ellipse (b/a)
            initial value : 1.
            min : 0.1
            max : 10.

    """

    def _set_units(self, x_unit, y_unit, z_unit, w_unit):

        # lon0 and lat0 and rdiff have most probably all units of degrees. However,
        # let's set them up here just to save for the possibility of using the
        # formula with other units (although it is probably never going to happen)

        self.lon0.unit = x_unit
        self.lat0.unit = y_unit
        self.rdiff0.unit = x_unit

        # Delta is of course unitless

        self.delta.unit = u.dimensionless_unscaled
        self.b.unit = u.dimensionless_unscaled
        self.incl.unit = x_unit
        self.elongation.unit = u.dimensionless_unscaled

        # Piv has the same unit as energy (which is z)

        self.piv.unit = z_unit
        self.piv2.unit = z_unit

    def evaluate(
        self, x, y, z, lon0, lat0, rdiff0, delta, b, piv, piv2, incl, elongation
    ):

        lon, lat = x, y
        energy = z

        # energy in kev -> TeV.
        # NOTE: the use of piv2 is necessary to preserve dimensional correctness: the logarithm can only be taken
        # of a dimensionless quantity, so there must be a pivot there.

        e_energy_piv2 = 17.0 * np.power(
            old_div(energy, piv2),
            0.54 + 0.046 * np.log10(old_div(energy, piv2)),
        )
        e_piv_piv2 = 17.0 * np.power(
            old_div(piv, piv2), 0.54 + 0.046 * np.log10(old_div(piv, piv2))
        )

        try:

            rdiff_a = (
                rdiff0
                * np.power(
                    old_div(e_energy_piv2, e_piv_piv2),
                    old_div((delta - 1.0), 2.0),
                )
                * np.sqrt(
                    b * b / 8.0 / np.pi * 0.624
                    + 0.26 * np.power(1.0 + 0.0107 * e_piv_piv2, -1.5)
                )
                / np.sqrt(
                    b * b / 8.0 / np.pi * 0.624
                    + 0.26 * np.power(1.0 + 0.0107 * e_energy_piv2, -1.5)
                )
            )

        except ValueError:

            # This happens when using units, because astropy.units fails with the message:
            # "ValueError: Quantities and Units may only be raised to a scalar power"

            # Work around the problem with this loop, which is slow but using units is only for testing purposes or
            # single calls, so it shouldn't matter too much
            rdiff_a = (
                np.array(
                    [
                        (rdiff0 * np.power(old_div(e_energy_piv2, e_piv_piv2), x)).value
                        for x in (delta - 1.0)
                        / 2.0
                        * np.sqrt(
                            b * b / 8.0 / np.pi * 0.624
                            + 0.26 * np.power(1.0 + 0.0107 * e_piv_piv2, -1.5)
                        )
                        / np.sqrt(
                            b * b / 8.0 / np.pi * 0.624
                            + 0.26 * np.power(1.0 + 0.0107 * e_energy_piv2, -1.5)
                        )
                    ]
                )
                * rdiff0.unit
            )

        rdiff_b = rdiff_a * elongation

        pi = np.pi

        angsep = angular_distance_fast(lon, lat, lon0, lat0)
        ang = np.arctan2(lat - lat0, (lon - lon0) * np.cos(lat0 * np.pi / 180.0))

        theta = np.arctan2(
            old_div(np.sin(ang - incl * np.pi / 180.0), elongation),
            np.cos(ang - incl * np.pi / 180.0),
        )

        rdiffs_a, thetas = np.meshgrid(rdiff_a, theta)
        rdiffs_b, angseps = np.meshgrid(rdiff_b, angsep)

        rdiffs = np.sqrt(
            rdiffs_a**2 * np.cos(thetas) ** 2 + rdiffs_b**2 * np.sin(thetas) ** 2
        )

        results = (
            np.power(old_div(180.0, pi), 2)
            * 1.22
            / (
                pi
                * np.sqrt(pi)
                * rdiffs_a
                * np.sqrt(elongation)
                * (angseps + 0.06 * rdiffs)
            )
            * np.exp(old_div(-np.power(angseps, 2), rdiffs**2))
        )

        return results

    def get_boundaries(self):

        # Truncate the function at the max of rdiff allowed

        maximum_rdiff = self.rdiff0.max_value

        min_latitude = max(-90.0, self.lat0.value - maximum_rdiff)
        max_latitude = min(90.0, self.lat0.value + maximum_rdiff)

        max_abs_lat = max(np.absolute(min_latitude), np.absolute(max_latitude))

        if (
            max_abs_lat > 89.0
            or old_div(maximum_rdiff, np.cos(max_abs_lat * np.pi / 180.0)) >= 180.0
        ):

            min_longitude = 0.0
            max_longitude = 360.0

        else:

            min_longitude = self.lon0.value - old_div(
                maximum_rdiff, np.cos(max_abs_lat * np.pi / 180.0)
            )
            max_longitude = self.lon0.value + old_div(
                maximum_rdiff, np.cos(max_abs_lat * np.pi / 180.0)
            )

            if min_longitude < 0.0:

                min_longitude += 360.0

            elif max_longitude > 360.0:

                max_longitude -= 360.0

        return (min_longitude, max_longitude), (min_latitude, max_latitude)

    def get_total_spatial_integral(self, z=None):
        """
        Returns the total integral (for 2D functions) or the integral over the spatial components (for 3D functions).
        needs to be implemented in subclasses.

        :return: an array of values of the integral (same dimension as z).
        """

        if isinstance(z, u.Quantity):
            z = z.value
        return np.ones_like(z)


class Continuous_injection_diffusion(Function3D, metaclass=FunctionMeta):
    r"""
    description :

        Positron and electrons diffusing away from the accelerator

    latex : $\left(\frac{180^\circ}{\pi}\right)^2 \frac{1.2154}{\sqrt{\pi^3} r_{\rm diff} ({\rm angsep} ({\rm x, y, lon_0, lat_0})+0.06 r_{\rm diff} )} \, {\rm exp}\left(-\frac{{\rm angsep}^2 ({\rm x, y, lon_0, lat_0})}{r_{\rm diff} ^2} \right)$

    parameters :

        lon0 :

            desc : Longitude of the center of the source
            initial value : 0.0
            min : 0.0
            max : 360.0

        lat0 :

            desc : Latitude of the center of the source
            initial value : 0.0
            min : -90.0
            max : 90.0

        rdiff0 :

            desc : Projected diffusion radius limited by the cooling time. The maximum allowed value is used to define the truncation radius.
            initial value : 1.0
            min : 0
            max : 20

        rinj :

            desc : Ratio of diffusion radius limited by the injection time over rdiff0. The maximum allowed value is used to define the truncation radius.
            initial value : 100.0
            min : 0
            max : 200
            fix : yes

        delta :

            desc : index for the diffusion coefficient
            initial value : 0.5
            min : 0.3
            max : 0.6
            fix : yes

        b :

            desc : b field strength in uG
            initial value : 3
            min : 1
            max : 10.
            fix : yes

        piv :

            desc : Pivot for the diffusion radius
            initial value : 2e10
            min : 0
            fix : yes

        piv2 :
            desc : Pivot for converting gamma energy to electron energy (always be 1 TeV)
            initial value : 1e9
            min : 0
            fix : yes

    """

    def _set_units(self, x_unit, y_unit, z_unit, w_unit):

        # lon0 and lat0 and rdiff have most probably all units of degrees. However,
        # let's set them up here just to save for the possibility of using the
        # formula with other units (although it is probably never going to happen)

        self.lon0.unit = x_unit
        self.lat0.unit = y_unit
        self.rdiff0.unit = x_unit
        self.rinj.unit = u.dimensionless_unscaled

        # Delta is of course unitless

        self.delta.unit = u.dimensionless_unscaled
        self.b.unit = u.dimensionless_unscaled

        # Piv has the same unit as energy (which is z)

        self.piv.unit = z_unit
        self.piv2.unit = z_unit

    def evaluate(self, x, y, z, lon0, lat0, rdiff0, rinj, delta, b, piv, piv2):

        lon, lat = x, y
        energy = z

        # energy in kev -> TeV.
        # NOTE: the use of piv2 is necessary to preserve dimensional correctness: the logarithm can only be taken
        # of a dimensionless quantity, so there must be a pivot there.

        e_energy_piv2 = 17.0 * np.power(
            old_div(energy, piv2),
            0.54 + 0.046 * np.log10(old_div(energy, piv2)),
        )
        e_piv_piv2 = 17.0 * np.power(
            old_div(piv, piv2), 0.54 + 0.046 * np.log10(old_div(piv, piv2))
        )

        rdiff_c = (
            rdiff0
            * np.power(old_div(e_energy_piv2, e_piv_piv2), old_div((delta - 1.0), 2.0))
            * np.sqrt(
                b * b / 8.0 / np.pi * 0.624
                + 0.26 * np.power(1.0 + 0.0107 * e_piv_piv2, -1.5)
            )
            / np.sqrt(
                b * b / 8.0 / np.pi * 0.624
                + 0.26 * np.power(1.0 + 0.0107 * e_energy_piv2, -1.5)
            )
        )

        rdiff_i = (
            rdiff0
            * rinj
            * np.power(old_div(e_energy_piv2, e_piv_piv2), old_div(delta, 2.0))
        )

        rdiff = np.minimum(rdiff_c, rdiff_i)

        angsep = angular_distance_fast(lon, lat, lon0, lat0)

        pi = np.pi

        rdiffs, angseps = np.meshgrid(rdiff, angsep)

        return (
            np.power(old_div(180.0, pi), 2)
            * 1.2154
            / (pi * np.sqrt(pi) * rdiffs * (angseps + 0.06 * rdiffs))
            * np.exp(old_div(-np.power(angseps, 2), rdiffs**2))
        )

    def get_boundaries(self):

        # Truncate the function at the max of rdiff allowed

        maximum_rdiff = self.rdiff0.max_value

        min_latitude = max(-90.0, self.lat0.value - maximum_rdiff)
        max_latitude = min(90.0, self.lat0.value + maximum_rdiff)

        max_abs_lat = max(np.absolute(min_latitude), np.absolute(max_latitude))

        if (
            max_abs_lat > 89.0
            or old_div(maximum_rdiff, np.cos(max_abs_lat * np.pi / 180.0)) >= 180.0
        ):

            min_longitude = 0.0
            max_longitude = 360.0

        else:

            min_longitude = self.lon0.value - old_div(
                maximum_rdiff, np.cos(max_abs_lat * np.pi / 180.0)
            )
            max_longitude = self.lon0.value + old_div(
                maximum_rdiff, np.cos(max_abs_lat * np.pi / 180.0)
            )

            if min_longitude < 0.0:

                min_longitude += 360.0

            elif max_longitude > 360.0:

                max_longitude -= 360.0

        return (min_longitude, max_longitude), (min_latitude, max_latitude)

    def get_total_spatial_integral(self, z=None):
        """
        Returns the total integral (for 2D functions) or the integral over the spatial components (for 3D functions).
        needs to be implemented in subclasses.

        :return: an array of values of the integral (same dimension as z).
        """

        if isinstance(z, u.Quantity):
            z = z.value
        return np.ones_like(z)


class Continuous_injection_diffusion_legacy(Function3D, metaclass=FunctionMeta):
    r"""
    description :

        Positron and electrons diffusing away from the accelerator

    latex : $\left(\frac{180^\circ}{\pi}\right)^2 \frac{1.2154}{\sqrt{\pi^3} r_{\rm diff} ({\rm angsep} ({\rm x, y, lon_0, lat_0})+0.06 r_{\rm diff} )} \, {\rm exp}\left(-\frac{{\rm angsep}^2 ({\rm x, y, lon_0, lat_0})}{r_{\rm diff} ^2} \right)$

    parameters :

        lon0 :

            desc : Longitude of the center of the source
            initial value : 0.0
            min : 0.0
            max : 360.0

        lat0 :

            desc : Latitude of the center of the source
            initial value : 0.0
            min : -90.0
            max : 90.0

        rdiff0 :

            desc : Projected diffusion radius. The maximum allowed value is used to define the truncation radius.
            initial value : 1.0
            min : 0
            max : 20

        delta :

            desc : index for the diffusion coefficient
            initial value : 0.5
            min : 0.3
            max : 0.6
            fix : yes

        uratio :

            desc : ratio between u_cmb and u_B
            initial value : 0.5
            min : 0.01
            max : 100.
            fix : yes

        piv :

            desc : Pivot for the diffusion radius
            initial value : 2e10
            min : 0
            fix : yes

        piv2 :
            desc : Pivot for converting gamma energy to electron energy (always be 1 TeV)
            initial value : 1e9
            min : 0
            fix : yes

    """

    def _set_units(self, x_unit, y_unit, z_unit, w_unit):

        # lon0 and lat0 and rdiff have most probably all units of degrees. However,
        # let's set them up here just to save for the possibility of using the
        # formula with other units (although it is probably never going to happen)

        self.lon0.unit = x_unit
        self.lat0.unit = y_unit
        self.rdiff0.unit = x_unit

        # Delta is of course unitless

        self.delta.unit = u.dimensionless_unscaled
        self.uratio.unit = u.dimensionless_unscaled

        # Piv has the same unit as energy (which is z)

        self.piv.unit = z_unit
        self.piv2.unit = z_unit

    def evaluate(self, x, y, z, lon0, lat0, rdiff0, delta, uratio, piv, piv2):

        lon, lat = x, y
        energy = z

        # energy in kev -> TeV.
        # NOTE: the use of piv2 is necessary to preserve dimensional correctness: the logarithm can only be taken
        # of a dimensionless quantity, so there must be a pivot there.

        e_energy_piv2 = 17.0 * np.power(
            old_div(energy, piv2),
            0.54 + 0.046 * np.log10(old_div(energy, piv2)),
        )
        e_piv_piv2 = 17.0 * np.power(
            old_div(piv, piv2), 0.54 + 0.046 * np.log10(old_div(piv, piv2))
        )

        try:

            rdiff = (
                rdiff0
                * np.power(
                    old_div(e_energy_piv2, e_piv_piv2),
                    old_div((delta - 1.0), 2.0),
                )
                * np.sqrt(1.0 + uratio * np.power(1.0 + 0.0107 * e_piv_piv2, -1.5))
                / np.sqrt(1.0 + uratio * np.power(1.0 + 0.0107 * e_energy_piv2, -1.5))
            )

        except ValueError:

            # This happens when using units, because astropy.units fails with the message:
            # "ValueError: Quantities and Units may only be raised to a scalar power"

            # Work around the problem with this loop, which is slow but using units is only for testing purposes or
            # single calls, so it shouldn't matter too much
            rdiff = (
                np.array(
                    [
                        (rdiff0 * np.power(old_div(e_energy_piv2, e_piv_piv2), x)).value
                        for x in (delta - 1.0)
                        / 2.0
                        * np.sqrt(
                            1.0 + uratio * np.power(1.0 + 0.0107 * e_piv_piv2, -1.5)
                        )
                        / np.sqrt(
                            1.0 + uratio * np.power(1.0 + 0.0107 * e_energy_piv2, -1.5)
                        )
                    ]
                )
                * rdiff0.unit
            )

        angsep = angular_distance_fast(lon, lat, lon0, lat0)

        pi = np.pi

        rdiffs, angseps = np.meshgrid(rdiff, angsep)

        return (
            np.power(old_div(180.0, pi), 2)
            * 1.2154
            / (pi * np.sqrt(pi) * rdiffs * (angseps + 0.06 * rdiffs))
            * np.exp(old_div(-np.power(angseps, 2), rdiffs**2))
        )

    def get_boundaries(self):

        # Truncate the function at the max of rdiff allowed

        maximum_rdiff = self.rdiff0.max_value

        min_latitude = max(-90.0, self.lat0.value - maximum_rdiff)
        max_latitude = min(90.0, self.lat0.value + maximum_rdiff)

        max_abs_lat = max(np.absolute(min_latitude), np.absolute(max_latitude))

        if (
            max_abs_lat > 89.0
            or old_div(maximum_rdiff, np.cos(max_abs_lat * np.pi / 180.0)) >= 180.0
        ):

            min_longitude = 0.0
            max_longitude = 360.0

        else:

            min_longitude = self.lon0.value - old_div(
                maximum_rdiff, np.cos(max_abs_lat * np.pi / 180.0)
            )
            max_longitude = self.lon0.value + old_div(
                maximum_rdiff, np.cos(max_abs_lat * np.pi / 180.0)
            )

            if min_longitude < 0.0:

                min_longitude += 360.0

            elif max_longitude > 360.0:

                max_longitude -= 360.0

        return (min_longitude, max_longitude), (min_latitude, max_latitude)

    def get_total_spatial_integral(self, z=None):
        """
        Returns the total integral (for 2D functions) or the integral over the spatial components (for 3D functions).
        needs to be implemented in subclasses.

        :return: an array of values of the integral (same dimension as z).
        """

        if isinstance(z, u.Quantity):
            z = z.value
        return np.ones_like(z)


class GalPropTemplate_3D(Function3D):
    r"""
    description :

        Use a 3D template that has morphology and flux information.
        GalProp, DRAGON or a similar model in fits format would work.
        Only parameter is a normalization factor.

    latex : $ K $

    parameters :

        K :

            desc : normalization
            initial value : 1
            fix : yes

        hash :

            desc : hash of model map [needed for memoization]
            initial value : 1
            fix : yes

    """

    __metaclass__ = FunctionMeta

    def _set_units(self, x_unit, y_unit, z_unit, w_unit):

        self.K.unit = (u.MeV * u.cm**2 * u.s * u.sr) ** (-1)

    def _setup(self):

        self._frame = ICRS()
        self._map = None
        self._fitsfile = None
        self._interpmap = None

    def set_frame(self, new_frame):
        """
        Set a new frame for the coordinates (the default is ICRS J2000)

        :param new_frame: a coordinate frame from astropy
        :return: (none)
        """
        assert isinstance(new_frame, BaseCoordinateFrame)

        self._frame = new_frame

    def load_file(self, fitsfile, phi1, phi2, theta1, theta2, galactic=False, ihdu=0):

        if fitsfile is None:
            raise RuntimeError("Need to specify a fits file with a template map.")

        self._fitsfile = fitsfile
        p1, p2, t1, t2 = self.define_region(phi1, phi2, theta1, theta2, galactic)
        self.ramin = p1
        self.ramax = p2
        self.decmin = t1
        self.decmax = t2

        with fits.open(self._fitsfile) as f:

            self._delLon = f[ihdu].header["CDELT1"]
            self._delLat = f[ihdu].header["CDELT2"]
            self._delEn = f[ihdu].header["CDELT3"]
            self._refLon = f[ihdu].header["CRVAL1"]
            self._refLat = f[ihdu].header["CRVAL2"]
            self._refEn = f[ihdu].header["CRVAL3"]  # values in log10
            self._map = f[ihdu].data
            self._nl = f[ihdu].header["NAXIS1"]  # longitude
            self._nb = f[ihdu].header["NAXIS2"]  # latitude
            self._ne = f[ihdu].header["NAXIS3"]  # energy

            # Create the function for the interpolation
            self._L = np.linspace(
                self._refLon,
                self._refLon + (self._nl - 1) * self._delLon,
                self._nl,
            )
            self._B = np.linspace(
                self._refLat,
                self._refLat + (self._nb - 1) * self._delLat,
                self._nb,
            )
            self._E = np.linspace(
                self._refEn,
                self._refEn + (self._ne - 1) * self._delEn,
                self._ne,
            )
            for i in range(len(self._E)):
                # Map units in Mev / cm^2 s sr, changing to 1 / MeV cm^2 s sr
                self._map[i] = self._map[i] / (
                    np.power(10, self._E[i]) * np.power(10, self._E[i])
                )
                self._map[i] = np.fliplr(self._map[i])
            self._F = RegularGridInterpolator(
                (self._E, self._B, self._L), self._map, bounds_error=False
            )

            h = hashlib.sha224()
            h.update(self._map)
            self.hash = int(h.hexdigest(), 16)

    def to_dict(self, minimal=False):

        data = super(Function3D, self).to_dict(minimal)

        if not minimal:

            data["extra_setup"] = {
                "_fitsfile": self._fitsfile,
                "_frame": self._frame,
                "ramin": self.ramin,
                "ramax": self.ramax,
                "decmin": self.decmin,
                "decmax": self.decmax,
            }

        return data

    def which_model_file(self):
        return self._fitsfile

    def evaluate(self, x, y, z, K, hash):

        if self._map is None:

            self.load_file(
                self._fitsfile,
                self.ramin,
                self.ramax,
                self.decmin,
                self.decmax,
                False,
                ihdu=0,
            )

        # Interpolated values can be cached since we are fitting the constant K
        if self._interpmap is None:

            # We assume x and y are R.A. and Dec
            _coord = SkyCoord(ra=x, dec=y, frame=self._frame, unit="deg")
            b = _coord.transform_to("galactic").b.value
            l = _coord.transform_to("galactic").l.value
            lon = l
            lat = b

            # transform energy from keV to MeV. Galprop Model starts at 100 MeV
            energy = np.log10(z) - np.log10((u.MeV.to("keV") / u.keV).value)

            if lon.size != lat.size:
                raise AttributeError("Lon and Lat should be the same size")
            f = np.zeros([lon.size, energy.size])
            E0 = self._refEn
            Ef = self._refEn + (self._ne - 1) * self._delEn

            # fix longitude
            shift = np.where(lon > 180.0)
            lon[shift] = 180 - lon[shift]

            for i in range(energy.size):
                e = np.repeat(energy[i], len(lon))
                try:
                    f[:, i] = self._F(zip(e, lat, lon))
                except ValueError:
                    pass
            bad_idx = np.isnan(f)
            f[bad_idx] = 0
            bad_idx = np.isinf(f)
            f[bad_idx] = 0
            assert np.all(np.isfinite(f)), "some interpolated values are wrong"
            self._interpmap = f

        # (1000 is to change from MeV to KeV)
        A = np.multiply(K, self._interpmap / 1000.0)
        return A

    def define_region(self, a, b, c, d, galactic=False):
        if galactic:
            lmin = a
            lmax = b
            bmin = c
            bmax = d

            _coord = SkyCoord(
                l=[lmin, lmin, lmax, lmax],
                b=[bmin, bmax, bmax, bmin],
                frame="galactic",
                unit="deg",
            )

            ramin = min(_coord.transform_to("icrs").ra.value)
            ramax = max(_coord.transform_to("icrs").ra.value)
            decmin = min(_coord.transform_to("icrs").dec.value)
            decmax = max(_coord.transform_to("icrs").dec.value)

        else:
            ramin = a
            ramax = b
            decmin = c
            decmax = d
        return ramin, ramax, decmin, decmax

    def get_boundaries(self):
        min_longitude = self.ramin
        max_longitude = self.ramax
        min_latitude = self.decmin
        max_latitude = self.decmax
        return (min_longitude, max_longitude), (min_latitude, max_latitude)



import numpy as np


def vincenty(lon0, lat0, a1, s):
    """
    Returns the coordinates of a new point that is a given angular distance s away from a starting point (lon0, lat0) at bearing (angle from north) a1), to within a given precision

    Note that this calculation is a simplified version of the full vincenty problem, which solves for the coordinates on the surface on an arbitrary ellipsoid. Here we only care about the surface of a sphere.

    Note: All parameters are assumed to be given in DEGREES
    :param lon0: float, longitude of starting point
    :param lat0: float, latitude of starting point
    :param a1: float, bearing to second point, i.e. angle between due north and line connecting 2 points
    :param s: float, angular distance between the two points
    :return: coordinates of second point in degrees
    """

    lon0 = np.deg2rad(lon0)
    lat0 = np.deg2rad(lat0)
    a1 = np.deg2rad(a1)
    s = np.deg2rad(s)

    sina = np.cos(lat0) * np.sin(a1)

    num1 = np.sin(lat0) * np.cos(s) + np.cos(lat0) * np.sin(s) * np.cos(a1)
    den1 = np.sqrt(
        sina**2 + (np.sin(lat0) * np.sin(s) - np.cos(lat0) * np.cos(a1)) ** 2
    )
    lat = np.rad2deg(np.arctan2(num1, den1))

    num2 = np.sin(s) * np.sin(a1)
    den2 = np.cos(lat0) * np.cos(s) - np.sin(lat0) * np.sin(s) * np.cos(a1)
    L = np.arctan2(num2, den2)
    lon = np.rad2deg(lon0 + L)

    return lon, lat



# If threeml is not installed, we create our own log,
# otherwise, just print to the 3ML one!

import logging
import logging.handlers as handlers
from contextlib import contextmanager
from pathlib import Path

from rich.console import Console
from rich.logging import RichHandler
from rich.theme import Theme

from astromodels.utils.configuration import astromodels_config

from .file_utils import _get_data_file_path

try:

    from threeML.config.config import threeML_config

    has_threeml = True

except ImportError:

    has_threeml = False


DEBUG_NODE_LEVEL = 9
logging.addLevelName(DEBUG_NODE_LEVEL, "DEBUG_NODE")


def debug_node(self, message, *args, **kws):
    if self.isEnabledFor(DEBUG_NODE_LEVEL):
        # Yes, logger takes its '*args' as 'args'.
        self._log(DEBUG_NODE_LEVEL, message, args, **kws)


logging.Logger.debug_node = debug_node


def get_path_of_log_dir():

    # we use the 3ML log path to simplify things
    # a more clever solution could be found

    if has_threeml:

        user_log: Path = Path(threeML_config.logging.path).expanduser()

    else:

        user_log: Path = Path(astromodels_config.logging.path).expanduser()

    # Create it if doesn't exist
    if not user_log.exists():

        user_log.mkdir(parents=True)

    return user_log


_log_file_names = ["usr.log", "dev.log"]


def get_path_of_log_file(log_file: str) -> Path:
    """
    returns the path of the log files
    """
    assert log_file in _log_file_names, f"{log_file} is not one of {_log_file_names}"

    return get_path_of_log_dir() / log_file


class LogFilter(object):
    def __init__(self, level):
        self.__level = level

    def filter(self, logRecord):
        return logRecord.levelno != self.__level


# now create the developer handler that rotates every day and keeps
# 10 days worth of backup
astromodels_dev_log_handler = handlers.TimedRotatingFileHandler(
    get_path_of_log_file("dev.log"), when="D", interval=1, backupCount=10
)

# lots of info written out

_dev_formatter = logging.Formatter(
    "%(asctime)s | %(name)s | %(levelname)s| %(funcName)s | %(lineno)d | %(message)s",
    datefmt="%Y-%m-%d %H:%M:%S",
)

astromodels_dev_log_handler.setFormatter(_dev_formatter)
astromodels_dev_log_handler.setLevel(logging.DEBUG)
# now set up the usr log which will save the info

astromodels_usr_log_handler = handlers.TimedRotatingFileHandler(
    get_path_of_log_file("usr.log"), when="D", interval=1, backupCount=10
)

astromodels_usr_log_handler.setLevel(logging.INFO)

# lots of info written out
_usr_formatter = logging.Formatter(
    "%(asctime)s | %(levelname)s | %(message)s", datefmt="%Y-%m-%d %H:%M:%S"
)

astromodels_usr_log_handler.setFormatter(_usr_formatter)


_console_formatter = logging.Formatter(
    " %(message)s",
    datefmt="%H:%M:%S",
)


_theme = {}

# Banner
_theme["h1"] = "deep_sky_blue3"
_theme["status.spinner"] = "cyan2"
_theme["status.text"] = "deep_sky_blue4"
_theme["repr.filename"] = "blue"
_theme["repr.number"] = "white"
_theme["repr.path"] = "grey37"
_theme["repr.str"] = "grey37"
_theme["repr.tag_name"] = "white"
_theme["repr.url"] = "not bold not italic underline grey84"
_theme["log.time"] = "green1"
_theme["log.message"] = f"{astromodels_config.logging.message_style}"
_theme["logging.level.debug"] = f"{astromodels_config.logging.debug_style}"
_theme["logging.level.error"] = f"{astromodels_config.logging.error_style}"
_theme["logging.level.info"] = f"{astromodels_config.logging.info_style}"
_theme["logging.level.warning"] = f"{astromodels_config.logging.warn_style}"
_theme["logging.level.degub_node"] = "light_goldenrod1"


# mytheme = Theme().read(_get_data_file_path("log_theme.ini"))
mytheme = Theme(_theme)
console = Console(theme=mytheme)


astromodels_console_log_handler = RichHandler(
    level="INFO", rich_tracebacks=True, markup=True, console=console
)
astromodels_console_log_handler.setFormatter(_console_formatter)
astromodels_console_log_handler.setLevel("INFO")

warning_filter = LogFilter(logging.WARNING)


@contextmanager
def silence_console_log():
    """
    temporarily silence the console and progress bars
    """
    current_console_logging_level = astromodels_console_log_handler.level
    current_usr_logging_level = astromodels_usr_log_handler.level

    astromodels_console_log_handler.setLevel(logging.CRITICAL)
    astromodels_usr_log_handler.setLevel(logging.CRITICAL)

    try:

        yield

    finally:

        astromodels_console_log_handler.setLevel(current_console_logging_level)
        astromodels_usr_log_handler.setLevel(current_usr_logging_level)


def silence_warnings():
    """
    supress warning messages in console and file usr logs
    """

    astromodels_usr_log_handler.addFilter(warning_filter)
    astromodels_console_log_handler.addFilter(warning_filter)


def activate_warnings():
    """
    supress warning messages in console and file usr logs
    """

    astromodels_usr_log_handler.removeFilter(warning_filter)
    astromodels_console_log_handler.removeFilter(warning_filter)


def update_logging_level(level):

    astromodels_console_log_handler.setLevel(level)


def setup_logger(name):

    # A logger with name name will be created
    # and then add it to the print stream
    log = logging.getLogger(name)

    # this must be set to allow debug messages through
    log.setLevel(DEBUG_NODE_LEVEL)

    # add the handlers

    log.addHandler(astromodels_dev_log_handler)

    log.addHandler(astromodels_console_log_handler)

    log.addHandler(astromodels_usr_log_handler)

    # we do not want to duplicate teh messages in the parents
    log.propagate = False

    return log



import math
from dataclasses import dataclass
from pathlib import Path
from typing import Any, Dict, List, Optional

import astropy.units as astropy_units
import numba as nb
import numpy as np
from astropy.io import fits
from interpolation import interp

from astromodels.functions.function import Function1D, FunctionMeta
from astromodels.utils import _get_data_file_path
from astromodels.utils.configuration import astromodels_config
from astromodels.utils.logging import setup_logger

log = setup_logger(__name__)

try:

    # ebltable is a Python packages to read in and interpolate tables for the photon density of
    # the Extragalactic Background Light (EBL) and the resulting opacity for high energy gamma
    # rays.

    import ebltable.tau_from_model as ebltau

    has_ebltable = True

except ImportError:

    if astromodels_config.logging.startup_warnings:

        log.warning(
            "The ebltable package is not available. Models that depend on it will not be available"
        )

    has_ebltable = False


class EBLTableNotAvailable(ImportWarning):
    pass


class InvalidUsageForFunction(Exception):
    pass


@dataclass(frozen=False)
class AbundanceTable:
    name: str
    tables: List[str]
    _current_table: str

    def set_table(self, table: str):

        """
        set the current table from AG89, WILM or ASPL

        :param table:
        :type table: str
        :returns:

        """

        old_table = self._current_table

        self._current_table = table

        if self.current_table not in self.tables:

            log.error(
                f"{self.name} does not contain {table} choose {','.join(self.table)}"
            )

            self._current_table = old_table

            raise AssertionError()

    @property
    def current_table(self) -> str:

        convert = {"AG89": "angr", "ASPL": "aspl", "WILM": "wilm"}

        return convert[self._current_table]

    @property
    def info(self) -> str:

        _abund_info = {}
        _abund_info[
            "WILM"
        ] = "wilms\nfrom Wilms, Allen & McCray (2000), ApJ 542, 914 \n except for elements not listed which are given zero abundance)\n https://heasarc.nasa.gov/xanadu/xspec/manual/XSabund.html "
        _abund_info[
            "AG89"
        ] = "angr\nfrom Anders E. & Grevesse N. (1989, Geochimica et Cosmochimica Acta 53, 197)\n https://heasarc.nasa.gov/xanadu/xspec/manual/XSabund.html"
        _abund_info[
            "ASPL"
        ] = "aspl\nfrom Asplund M., Grevesse N., Sauval A.J. & Scott P. (2009, ARAA, 47, 481)\nhttps://heasarc.nasa.gov/xanadu/xspec/manual/XSabund.html"

        return _abund_info[self._current_table]

    @property
    def xsect_table(self) -> np.ndarray:

        """
        returns the XSECT table for the current model

        :returns:

        """
        _path: Path = Path("xsect") / f"xsect_{self.name}_{self.current_table}.fits"

        path_to_xsect: Path = _get_data_file_path(_path)

        with fits.open(path_to_xsect) as fxs:

            dxs = fxs[1].data
            xsect_ene = dxs["ENERGY"]
            xsect_val = dxs["SIGMA"]

        return np.array(xsect_ene, dtype=np.float64), np.array(
            xsect_val, dtype=np.float64
        )


phabs = AbundanceTable("phabs", ["angr", "aspl"], "AG89")
tbabs = AbundanceTable("tbabs", ["angr", "aspl", "wilm"], "WILM")
wabs = AbundanceTable("wabs", ["angr"], "AG89")


# PhAbs class


class PhAbs(Function1D, metaclass=FunctionMeta):
    r"""
    description :
        Photometric absorption (phabs implementation), f(E) = exp(- NH * sigma(E))
        contributed by Dominique Eckert
    parameters :
        NH :
            desc : absorbing column density in units of 1e22 particles per cm^2
            initial value : 1.0
            is_normalization : False
            transformation : log10
            min : 1e-4
            max : 1e4
            delta : 0.1

        redshift :
            desc : the redshift of the source
            initial value : 0.
            is_normalization : False
            min : 0
            max : 15
            delta : 0.1
            fix: True

    properties:
         abundance_table:
            desc: the abundance table for the model
            initial value: AG89
            allowed values:
            - AG89
            - ASPL
            function: _init_xsect

    """

    def _setup(self):
        self._fixed_units = (
            astropy_units.keV,
            astropy_units.dimensionless_unscaled,
        )

        # # astromodels_config.absorption_models.phabs_table.value
        # self.init_xsect(self.abundance_table.value)

    def _set_units(self, x_unit, y_unit):
        self.NH.unit = astropy_units.cm ** (-2)
        self.redshift.unit = astropy_units.dimensionless_unscaled

    def _init_xsect(self):
        """
        Set the abundance table

        :param abund_table: "ASPL", "AG89"
        :returns:
        :rtype:

        """

        # load cross section data

        phabs.set_table(self.abundance_table.value)

        self.xsect_ene, self.xsect_val = phabs.xsect_table

    @property
    def abundance_table_info(self):
        print(phabs.info)

    def evaluate(self, x, NH, redshift):

        if isinstance(x, astropy_units.Quantity):

            _unit = astropy_units.cm**2
            _y_unit = astropy_units.dimensionless_unscaled
            _x = x.value
            _redshift = redshift.value
        else:

            _unit = 1.0
            _y_unit = 1.0
            _redshift = redshift
            _x = x

        xsect_interp = interp(self.xsect_ene, self.xsect_val, _x * (1 + _redshift))

        # evaluate the exponential with numba

        spec = _numba_eval(NH, xsect_interp) * _y_unit

        return spec


# TbAbs class


class TbAbs(Function1D, metaclass=FunctionMeta):
    r"""
    description :
        Photometric absorption (Tbabs implementation), f(E) = exp(- NH * sigma(E))
        contributed by Dominique Eckert
    parameters :
        NH :
            desc : absorbing column density in units of 1e22 particles per cm^2
            initial value : 1.0
            is_normalization : True
            transformation : log10
            min : 1e-4
            max : 1e4
            delta : 0.1

        redshift :
            desc : the redshift of the source
            initial value : 0.
            is_normalization : False
            min : 0
            max : 15
            delta : 0.1
            fix: True

    properties:
         abundance_table:
            desc: the abundance table for the model
            initial value: WILM
            allowed values:
            - WILM
            - AG89
            - ASPL
            function: _init_xsect

    """

    def _setup(self):

        # self.init_xsect(self.abundance_table)

        self._fixed_units = (
            astropy_units.keV,
            astropy_units.dimensionless_unscaled,
        )

    def _set_units(self, x_unit, y_unit):

        self.NH.unit = astropy_units.cm ** (-2)
        self.redshift.unit = astropy_units.dimensionless_unscaled

    def _init_xsect(self):
        """
        Set the abundance table

        :param abund_table: "WILM", "ASPL", "AG89"
        :returns:
        :rtype:

        """

        tbabs.set_table(self.abundance_table.value)

        self.xsect_ene, self.xsect_val = tbabs.xsect_table

        log.debug(f"updated the TbAbs table to {self.abundance_table.value}")

    @property
    def abundance_table_info(self):
        print(tbabs.info)

    def evaluate(self, x, NH, redshift):

        if isinstance(x, astropy_units.Quantity):

            _unit = astropy_units.cm**2
            _y_unit = astropy_units.dimensionless_unscaled
            _x = x.value
            _redshift = redshift.value
        else:

            _unit = 1.0
            _y_unit = 1.0
            _redshift = redshift
            _x = x

        xsect_interp = interp(self.xsect_ene, self.xsect_val, _x * (1 + _redshift))

        spec = _numba_eval(NH, xsect_interp) * _y_unit

        return spec


# WAbs class


class WAbs(Function1D, metaclass=FunctionMeta):
    r"""
    description :
        Photometric absorption (Wabs implementation), f(E) = exp(- NH * sigma(E))
        contributed by Dominique Eckert
    parameters :
        NH :
            desc : absorbing column density in units of 1e22 particles per cm^2
            initial value : 1.0
            is_normalization : True
            transformation : log10
            min : 1e-4
            max : 1e4
            delta : 0.1
        redshift :
            desc : the redshift of the source
            initial value : 0.
            is_normalization : False
            min : 0
            max : 15
            delta : 0.1
            fix: True

    """

    def _setup(self):
        self._fixed_units = (
            astropy_units.keV,
            astropy_units.dimensionless_unscaled,
        )
        self._init_xsect()

    def _set_units(self, x_unit, y_unit):
        self.NH.unit = astropy_units.cm ** (-2)
        self.redshift.unit = astropy_units.dimensionless_unscaled

    def _init_xsect(self):
        """
        Set the abundance table

        :returns:
        :rtype:

        """

        self.xsect_ene, self.xsect_val = wabs.xsect_table

    @property
    def abundance_table_info(self):
        print(wabs.info)

    def evaluate(self, x, NH, redshift):

        if isinstance(x, astropy_units.Quantity):

            _unit = astropy_units.cm**2
            _y_unit = astropy_units.dimensionless_unscaled
            _x = x.value
            _redshift = redshift.value
        else:

            _unit = 1.0
            _y_unit = 1.0
            _redshift = redshift
            _x = x

        xsect_interp = interp(self.xsect_ene, self.xsect_val, _x * (1 + _redshift))

        spec = _numba_eval(NH, xsect_interp) * _y_unit

        return spec


if has_ebltable:

    class EBLattenuation(Function1D, metaclass=FunctionMeta):
        r"""
        description :
            Attenuation factor for absorption in the extragalactic background light (EBL) ,
            to be used for extragalactic source spectra. Based on package "ebltable" by
            Manuel Meyer, https://github.com/me-manu/ebltable .

        latex: not available

        parameters :

          redshift :
                desc : redshift of the source
                initial value : 1.0
                fix : yes

          attenuation :
                desc : scaling factor for the strength of attenuation
                initial value : 1.0
                min : 0.0
                max : 10.0
                fix : yes

        properties:
           ebl_model:
              desc: set the EBL model
              initial value: dominguez
              allowed values:
                 - dominguez
                 - franceschini
                 - kneiske
                 - inuoe
                 - gilmore
              function: _set_ebl_model


        """

        # def _setup(self):

        #     # define EBL model, use dominguez as default
        #     self._tau = ebltau.OptDepth.readmodel(model=astromodels_config.absorption_models.ebl_table.value)

        def _set_ebl_model(self):

            # passing modelname to ebltable, which will check if defined
            self._tau = ebltau.OptDepth.readmodel(model=self.ebl_model.value)

        def _set_units(self, x_unit, y_unit):

            if not hasattr(x_unit, "physical_type") or x_unit.physical_type != "energy":

                # x should be energy
                raise InvalidUsageForFunction(
                    "Unit for x is not an energy. The function "
                    "EBLOptDepth calculates energy-dependent "
                    "absorption."
                )

            # y should be dimensionless
            if (
                not hasattr(y_unit, "physical_type")
                or y_unit.physical_type != "dimensionless"
            ):
                raise InvalidUsageForFunction("Unit for y is not dimensionless.")

            self.redshift.unit = astropy_units.dimensionless_unscaled
            self.attenuation.unit = astropy_units.dimensionless_unscaled

        def evaluate(self, x, redshift, attenuation):

            if isinstance(x, astropy_units.Quantity):

                # ebltable expects TeV
                eTeV = x.to(astropy_units.TeV).value
                _unit = astropy_units.dimensionless_unscaled
                _redshift = redshift.value
                _attenuation = attenuation.value

            else:

                # otherwise it's in keV
                eTeV = x / 1.0e9

                _unit = 1.0
                _redshift = redshift
                _attenuation = attenuation

            return (
                _numba_eval(self._tau.opt_depth(_redshift, eTeV), _attenuation) * _unit
            )


@nb.vectorize
def _exp(x):
    return math.exp(x)


@nb.njit(fastmath=True)
def _numba_eval(nh, xsect_interp):

    return _exp(-nh * xsect_interp)



from .functions_1D import (
    Band,
    Band_Calderone,
    Band_grbm,
    Blackbody,
    Broken_powerlaw,
    Constant,
    Cubic,
    Cutoff_powerlaw,
    Cutoff_powerlaw_Ep,
    DiracDelta,
    Exponential_cutoff,
    Inverse_cutoff_powerlaw,
    Line,
    Log_parabola,
    ModifiedBlackbody,
    NonDissipativePhotosphere,
    NonDissipativePhotosphere_Deep,
    PhAbs,
    Powerlaw,
    Powerlaw_Eflux,
    Powerlaw_flux,
    Quadratic,
    Quartic,
    Sin,
    SmoothlyBrokenPowerLaw,
    DoubleSmoothlyBrokenPowerlaw,
    Standard_Rv,
    StepFunction,
    StepFunctionUpper,
    GenericFunction,
    Super_cutoff_powerlaw,
    TbAbs,
    WAbs,
    ZDust,
    get_polynomial,
    has_atomdb,
    has_ebltable,
    has_gsl,
    has_naima,
)

if has_naima:
    from .functions_1D import Synchrotron

if has_gsl:
    from .functions_1D import Cutoff_powerlaw_flux

if has_ebltable:
    from .functions_1D import EBLattenuation

if has_atomdb:

    from .functions_1D import APEC, VAPEC

from .dark_matter.dm_models import DMFitFunction, DMSpectra
from .function import (
    Function1D,
    Function2D,
    Function3D,
    FunctionMeta,
    ModelAssertionViolation,
)
from .functions_2D import (
    Asymm_Gaussian_on_sphere,
    Disk_on_sphere,
    Ellipse_on_sphere,
    Gaussian_on_sphere,
    Latitude_galactic_diffuse,
    Power_law_on_sphere,
    SpatialTemplate_2D,
)
from .functions_3D import (
    Continuous_injection_diffusion,
    Continuous_injection_diffusion_ellipse,
    Continuous_injection_diffusion_legacy,
    GalPropTemplate_3D,
)
from .priors import (
    Cauchy,
    Cosine_Prior,
    Gaussian,
    Log_normal,
    Log_uniform_prior,
    Truncated_gaussian,
    Uniform_prior,
    Beta,
    Gamma,
    Exponential,
    Powerlaw_Prior,
)
from .template_model import (
    MissingDataFile,
    TemplateModel,
    TemplateModelFactory,
    XSPECTableModel,
)

__all__ = [
    "Band",
    "Band_Calderone",
    "Band_grbm",
    "Broken_powerlaw",
    "Cutoff_powerlaw",
    "Cutoff_powerlaw_Ep",
    "Inverse_cutoff_powerlaw",
    "Powerlaw",
    "Powerlaw_Eflux",
    "Powerlaw_flux",
    "SmoothlyBrokenPowerLaw",
    "DoubleSmoothlyBrokenPowerlaw",
    "Super_cutoff_powerlaw",
    "Constant",
    "Cubic",
    "DiracDelta",
    "Exponential_cutoff",
    "Line",
    "Quadratic",
    "Sin",
    "StepFunction",
    "StepFunctionUpper",
    "GenericFunction",
    "PhAbs",
    "TbAbs",
    "WAbs",
    "Asymm_Gaussian_on_sphere",
    "Disk_on_sphere",
    "Ellipse_on_sphere",
    "Gaussian_on_sphere",
    "Latitude_galactic_diffuse",
    "Power_law_on_sphere",
    "SpatialTemplate_2D",
    "Continuous_injection_diffusion",
    "Continuous_injection_diffusion_ellipse",
    "Continuous_injection_diffusion_legacy",
    "GalPropTemplate_3D",
    "DMSpectra",
    "DMFitFunction",
    "Cauchy",
    "Cosine_Prior",
    "Gaussian",
    "Log_normal",
    "Log_uniform_prior",
    "Truncated_gaussian",
    "Uniform_prior",
    "Beta",
    "Gamma",
    "Exponential",
    "Powerlaw_Prior",
    "TemplateModel",
    "TemplateModelFactory",
    "XSPECTableModel",
    "MissingDataFile",
    "Log_parabola",
    "Blackbody",
    "Function1D",
    "Function2D",
    "Function3D",
    "FunctionMeta",
    "ModelAssertionViolation",
    "Quartic",
    "get_polynomial",
    "ZDust",
    "Standard_Rv",
    "ModifiedBlackbody",
    "NonDissipativePhotosphere",
    "NonDissipativePhotosphere_Deep",
]

if has_atomdb:
    __all__.extend(["APEC", "VAPEC"])

if has_gsl:

    __all__.extend(["Cutoff_powerlaw_flux"])

if has_naima:

    __all__.extend(["Synchrotron"])

if has_ebltable:

    __all__.extend(["EBLattenuation"])



from __future__ import division

import os

import copy

import numpy as np
import math

from astromodels import u, update_logging_level
from astromodels.core.polarization import *
from astromodels.core.model import Model
from astromodels.core.model_parser import *
from astromodels.functions import Constant,Powerlaw
from astromodels.sources.point_source import PointSource

update_logging_level("DEBUG")

def test_linear_polarization_parameters():
    degree = 50.0
    angle = 30.0
    ps = PointSource('PS',0,0,spectral_shape=Powerlaw(),polarization=LinearPolarization(degree=degree, angle=angle))
    m1 = Model(ps)
    m1.display()

    m1.save("__test.yml",overwrite=True)

    mp = load_model("__test.yml")
    assert math.isclose(mp.sources["PS"].spectrum.main.polarization.degree.value, degree, rel_tol=0.02)
    assert math.isclose(mp.sources["PS"].spectrum.main.polarization.angle.value, angle, rel_tol=0.02)
    mp.display()

    os.remove("__test.yml")

def test_linear_polarization_functions():
    degree = Constant()
    angle  = Constant()
    degree.k = 50
    angle.k  = 30
    ps = PointSource('PS',0,0,spectral_shape=Powerlaw(),polarization=LinearPolarization(degree=degree, angle=angle))
    m1 = Model(ps)
    m1.display()

    m1.save("__test.yml",overwrite=True)

    mp = load_model("__test.yml")
    assert math.isclose(mp.sources["PS"].spectrum.main.polarization.degree.Constant.k.value, degree.k.value, rel_tol=0.02)
    assert math.isclose(mp.sources["PS"].spectrum.main.polarization.angle.Constant.k.value, angle.k.value, rel_tol=0.02)
    mp.display()

    os.remove("__test.yml")

def test_Stokes_polarization_functions():
    u = Constant()
    q = Constant()
    u.k = 0.5
    q.k = 0.5

    ps = PointSource('PS',0,0,spectral_shape=Powerlaw(),polarization=StokesPolarization(Q=q, U=u))
    m1 = Model(ps)
    m1.display()

    m1.save("__test.yml",overwrite=True)

    mp = load_model("__test.yml")
    assert math.isclose(mp.sources["PS"].spectrum.main.polarization.Q.Constant.k.value, q.k.value, rel_tol=0.02)
    assert math.isclose(mp.sources["PS"].spectrum.main.polarization.Q.Constant.k.value, u.k.value, rel_tol=0.02)
    mp.display()

    os.remove("__test.yml")


import collections
import copy
from typing import Any, Dict, List, Optional, Tuple, Union

from astromodels.utils.logging import setup_logger

from .tree import Node

log = setup_logger(__name__)

# Exception for when a parameter is out of its bounds


class SettingUnknownValue(RuntimeError):
    pass


class PropertyBase(Node):
    def __init__(
        self,
        name: str,
        desc: str,
        value: Optional[str] = None,
        allowed_values: Optional[List[str]] = None,
        defer: bool = False,
        eval_func: Optional[str] = None,
    ):
        # Make this a node

        Node.__init__(self, name)

        self._allowed_values: Optional[List[str]] = allowed_values
        self._defer: bool = defer
        self._eval_func: Optional[str] = eval_func

        if (value is None) and (not self._defer):
            log.error(
                f"property {name} was given no initial value but is NOT deferred"
            )

        # now we set the value

        self.value = value

        self.__doc__ = desc

        self._desc = desc

    def _get_value(self) -> Any:
        """
        Return current parameter value
        """

        log.debug_node(
            f"accessing the property {self.name} with value {self._internal_value}"
        )

        return self._internal_value

    def _set_value(self, new_value) -> None:
        """
        Sets the current value of the parameter, ensuring that it is within the allowed range.
        """

        if (self._defer) and (new_value is None):
            # this is ok
            pass

        elif self._allowed_values is not None:
            if new_value not in self._allowed_values:
                log.error(
                    f"{self.name} can only take the values {','.join(self._allowed_values)} not {new_value}"
                )

                raise SettingUnknownValue()

        self._internal_value = new_value

        # if there is an eval func value
        # then we need to execute the function
        # on the parent

        if (self._internal_value == "_tmp") and self._defer:
            # do not execute in this mode

            return

        if self._eval_func is not None:
            # if there is a parent
            if self._parent is not None:
                if self._parent.name == "composite":
                    # ok, we have a composite function

                    func_idx = int(self._name.split("_")[-1]) - 1

                    log.debug_node(f"{self._name} has a composite parent and")
                    log.debug_node(f"is being executed on func idx {func_idx}")
                    log.debug_node(
                        f"and the parent has {len(self._parent._functions)} functions"
                    )

                    getattr(
                        self._parent._functions[func_idx], str(self._eval_func)
                    )()

                else:
                    getattr(self._parent, str(self._eval_func))()

            # other wise this will run when the parent is set

    value = property(
        _get_value,
        _set_value,
        doc="Get and sets the current value for the property",
    )

    def _set_parent(self, parent):
        # we intecept here becuase we want
        # to make sure the eval works

        super(PropertyBase, self)._set_parent(parent)

        # now we want to update because we have a parent
        self.value = self._internal_value

    @property
    def is_deferred(self) -> bool:
        return self._defer

    @property
    def description(self) -> Optional[str]:
        """
        Return a description of this parameter

        :return: a string cointaining a description of the meaning of this parameter
        """
        return self._desc

    def duplicate(self) -> "FunctionProperty":
        """
        Returns an exact copy of the current property
        """

        # Deep copy everything to make sure that there are no ties between the new instance and the old one

        new_property = copy.deepcopy(self)

        return new_property

    def _repr__base(self, rich_output):  # pragma: no cover
        raise NotImplementedError(
            "You need to implement this for the actual Property class"
        )

    @staticmethod
    def _to_python_type(variable):
        """
        Returns the value in the variable handling also np.array of one element

        :param variable: input variable
        :return: the value of the variable having a python type (int, float, ...)
        """

        # Assume variable is a np.array, fall back to the case where variable is already a primitive type

        try:
            return variable.item()

        except AttributeError:
            return variable

    def to_dict(self, minimal=False) -> Dict[str, Any]:
        """Returns the representation for serialization"""

        data = collections.OrderedDict()

        if minimal:
            # In the minimal representation we just output the value

            data["value"] = self._to_python_type(self.value)

        else:
            # In the complete representation we output everything is needed to re-build the object

            data["value"] = (
                self.value if type(self.value) is bool else str(self.value)
            )
            data["desc"] = str(self._desc)
            data["allowed values"] = self._to_python_type(self._allowed_values)
            data["defer"] = self._to_python_type(self._defer)
            data["function"] = str(self._eval_func)

        return data


class FunctionProperty(PropertyBase):
    def __init__(
        self,
        name: str,
        desc: str,
        value: Optional[str] = None,
        allowed_values: Optional[List[Any]] = None,
        defer: bool = False,
        eval_func: Optional[str] = None,
    ):
        super(FunctionProperty, self).__init__(
            name=name,
            desc=desc,
            value=value,
            allowed_values=allowed_values,
            defer=defer,
            eval_func=eval_func,
        )

    def _repr__base(self, rich_output=False):
        representation = (
            f"Property {self.name} = {self.value}\n"
            f"(allowed values = {'all' if self._allowed_values is None else ' ,'.join(self._allowed_values)})"
        )

        return representation



__author__ = "giacomov"


from astromodels.core.tree import Node
from astromodels.core.parameter import Parameter
import numpy as np


class Polarization(Node):
    def __init__(self, polarization_type="linear"):

        assert polarization_type in [
            "linear",
            "stokes",
        ], "polarization must be linear or stokes"

        self._polarization_type = polarization_type

        Node.__init__(self, "polarization")

    @staticmethod
    def _get_parameter_from_input(
        number_or_parameter, minimum, maximum, what, desc, unit
    ):

        # Try to transform it to float, if it works than we transform it to a parameter

        try:

            number_or_parameter = float(number_or_parameter)

        except TypeError:

            assert isinstance(number_or_parameter, Parameter), (
                "%s must be either a number or a " "parameter instance" % what
            )

            # So this is a Parameter instance already. Enforce that it has the right maximum and minimum

            parameter = number_or_parameter

            assert parameter.min_value == minimum, "%s must have a minimum of %s" % (
                what,
                minimum,
            )
            assert parameter.max_value == maximum, "%s must have a maximum of %s" % (
                what,
                maximum,
            )

        else:

            # This was a float. Enforce that it has a legal value

            assert (
                minimum <= number_or_parameter <= maximum
            ), "%s cannot have a value of %s, " "it must be %s <= %s <= %s" % (
                what,
                number_or_parameter,
                minimum,
                what,
                maximum,
            )

            parameter = Parameter(
                what,
                number_or_parameter,
                desc=desc,
                min_value=minimum,
                max_value=maximum,
                unit=unit,
                free=True,
            )

        return parameter


# TODO: add transform between polarizations

class LinearPolarization(Polarization):
    def __init__(self, degree, angle):
        """
        Linear parameterization of polarization

        :param degree: The polarization degree
        :param angle: The polarization angle
        """
        super(LinearPolarization, self).__init__(polarization_type="linear")

        if callable(degree):
            self.degree = LinearParameter('degree', degree)
        else:
            self.degree = self._get_parameter_from_input(degree, 0, 100, 'degree', 'Polarization degree', 'dimensionless_unscaled')

        if callable(angle):
            self.angle = LinearParameter('angle', angle)
        else:
            self.angle = self._get_parameter_from_input(angle, 0, 180, 'angle', 'Polarization angle', 'deg')

        self._add_child(self.degree)
        self._add_child(self.angle)

    def __call__(self, energies, stokes):

        if stokes == 'Q':
            return self.degree(energies) * np.cos(2.0 * np.radians(self.angle(energies)))
        elif stokes == 'U':
            return self.degree(energies) * np.sin(2.0 * np.radians(self.angle(energies)))
        return 1


class StokesPolarization(Polarization):
    def __init__(self, I=None, Q=None, U=None, V=None):
        """
        Stokes parameterization of polarization
        """
        super(StokesPolarization, self).__init__(polarization_type='stokes')
        self._Q = StokesParameter('Q', Q)
        self._add_child(self._Q)
        self._U = StokesParameter('U', U)
        self._add_child(self._U)

    def __call__(self, energies, stokes):
        if stokes == 'Q':
            return self._Q(energies)
        elif stokes == 'U':
            return self._U(energies)
        return 1

#     def to_linear_polarization(self):
#         # polarization angle
# #        psi = 0.5 * np.arctan2(U_bin, Q_bin)
#
#         # polarization fraction
# #        frac = np.sqrt(Q_bin ** 2 + U_bin ** 2) / I_bin
#
#         pass
#
#         #angle = 0.5 * np.arctan2(se)
#
#


class StokesParameter(Node):

    def __init__(self, name, value):

        assert name in ['I', 'Q', 'U', 'V']
        Node.__init__(self, name)
        self._add_child(value)
        self.value = value

    def __call__(self, energies):
        return self.value(energies)


class LinearParameter(Node):
    def __init__(self, name, value):
        assert name in ['degree', 'angle']
        Node.__init__(self, name)
        self._add_child(value)
        self.value = value

    def __call__(self, energies):
        return self.value(energies)



from pathlib import Path
from typing import Any, Dict, Optional

from omegaconf import OmegaConf
from omegaconf.dictconfig import DictConfig
from rich.tree import Tree

from astromodels.utils import get_path_of_user_config

from .config_structure import Config

# Read the default Config
astromodels_config: Config = OmegaConf.structured(Config)

# now glob the config directory

for user_config_file in get_path_of_user_config().glob("*.yml"):

    _partial_conf = OmegaConf.load(user_config_file)

    astromodels_config: Config = OmegaConf.merge(astromodels_config, _partial_conf)


def recurse_dict(d, tree):

    for k, v in d.items():

        if (type(v) is dict) or isinstance(v, DictConfig):

            branch = tree.add(
                k, guide_style="bold medium_orchid", style="bold medium_orchid"
            )

            recurse_dict(v, branch)

        else:

            tree.add(
                f"{k}: [blink cornflower_blue]{v}",
                guide_style="medium_spring_green",
                style="medium_spring_green",
            )

    return


def show_configuration(sub_menu: Optional[str] = None):
    """
    display the current configuration or a sub menu if
    provided
    """

    if sub_menu is None:

        tree = Tree(
            "config",
            guide_style="bold medium_orchid",
            style="bold medium_orchid",
        )

        recurse_dict(astromodels_config, tree)

    else:

        if sub_menu in astromodels_config:

            tree = Tree(
                "config",
                guide_style="bold medium_orchid",
                style="bold medium_orchid",
            )

            recurse_dict(astromodels_config[sub_menu], tree)

        else:

            msg = f"{sub_menu} is not in the astromodels configuration"

            raise AssertionError(msg)

    return tree



import math
from dataclasses import dataclass
from enum import Enum
from typing import Dict

import astropy.units as astropy_units
import numba as nb
import numpy as np

from astromodels.functions.function import Function1D, FunctionMeta
from astromodels.utils.logging import setup_logger

log = setup_logger(__name__)


@dataclass(frozen=True)
class _ExtinctionCurve:
    """
    simple extinction container
    """

    a: np.ndarray
    lamb: np.ndarray
    b: np.ndarray
    n: np.ndarray


class ZDust(Function1D, metaclass=FunctionMeta):
    r"""
    description :
        Extinction by dust grains from Pei (1992), suitable for IR, optical and UV energy bands, including the full energy ranges of the Swift UVOT and XMM-Newton OM detectors. Three models are included which characterize the extinction curves of (1) the Milky Way, (2) the LMC and (3) the SMC. The models can be modified by redshift and can therefore be applied to extragalactic sources. The transmission is set to unity shortward of 912 Angstroms in the rest frame of the dust. This is incorrect physically but does allow the model to be used in combination with an X-ray photoelectric absorption model such as phabs. Parameter 1 (method) describes which extinction curve (MW, LMC or SMC) will be constructed and should never be allowed to float during a fit. The extinction at V, A(V) = E(B-V) x Rv. Rv should typically remain frozen for a fit. Standard values for Rv are MW = 3.08, LMC = 3.16 and SMC = 2.93 (from table 2 of Pei 1992), although these may not be applicable to more distant dusty sources.
    parameters :
        e_bmv :
            desc : color excess
            initial value : 1.0
            min: 0
            is_normalization : False
            delta : 0.1

        rv :
            desc :  ratio of total to selective extinction
            initial value : 3.08
            is_normalization : False
            delta : 0.1
            fix: True

        redshift :
            desc : the redshift of the source
            initial value : 0.
            is_normalization : False
            min : 0
            max : 15
            delta : 0.1
            fix: True

    properties:
        extinction_law:
            desc: the abundance table for the model
            initial value: mw
            allowed values:
            - mw
            - lmc
            - smc

            function: _set_extinction_law



    """

    def _setup(self):
        self._fixed_units = (
            astropy_units.keV,
            astropy_units.dimensionless_unscaled,
        )

    def _set_units(self, x_unit, y_unit):
        self.rv.unit = astropy_units.dimensionless_unscaled
        self.e_bmv.unit = astropy_units.dimensionless_unscaled
        self.redshift.unit = astropy_units.dimensionless_unscaled

    def _set_extinction_law(self) -> None:

        self._extinction_law: _ExtinctionCurve = _extinctions_laws[
            self.extinction_law.value
        ]

    def evaluate(self, x, e_bmv, rv, redshift):

        if isinstance(x, astropy_units.Quantity):

            _x = np.array(x.to("keV").value, ndmin=1, copy=False, dtype=float)

            _unit = astropy_units.cm**2
            _y_unit = astropy_units.dimensionless_unscaled

            _e_bmv = e_bmv.value
            _rv = rv.value
            _redshift = redshift.value

        else:

            _unit = 1.0
            _y_unit = 1.0
            _redshift = redshift
            _e_bmv = e_bmv
            _rv = rv
            _x = x

        # _x = np.append(_x, _x[-1]* 1.01)

        zfac = _redshift + 1.0

        return (
            ms_dust(
                _x * zfac,
                _e_bmv,
                _rv,
                self._extinction_law.a,
                self._extinction_law.lamb,
                self._extinction_law.b,
                self._extinction_law.n,
            )
            * _y_unit
        )


@nb.njit(fastmath=True)
def ms_dust(x, e_bmv, rv, a, lamb, b, n):

    # conversion constant in keV*\AA

    hc = 12.3963

    # extinction at B (a_b)

    a_b = e_bmv * (1 + rv)

    ne = len(x)

    xx = hc / x

    out = np.zeros(ne)

    for i in range(ne):

        out[i] = pei(xx[i], a_b, a, lamb, b, n)

    return out


@nb.njit(fastmath=True)
def ms_dust_xspec(x, e_bmv, rv, a, lamb, b, n):

    # conversion constant in keV*\AA

    hc = 12.3963

    #  photon index of the base power law function
    index = 2.0

    # extinction at B (a_b)

    a_b = rv * (1 + e_bmv)

    ne = len(x)

    out = np.zeros(ne)

    xx = hc / x

    xl = xx[0]
    sl = math.pow(xl, -index)
    fl = pei(xl, a_b, a, lamb, b, n)

    for i in range(ne):
        xh = xx[i + 1]
        sh = math.pow(xh, -index)
        fh = pei(xh, a_b, a, lamb, b, n)

        out[i] = (sl * fl + sh * fh) / (sl + sh)

        xl = xh
        sl = sh
        fl = fh

    return out


@nb.njit(fastmath=True)
def pei(rlambda, a_b, a, lamb, b, n) -> float:
    """
    ported from XSPEC originally by
    Martin.Still@gsfc.nasa.gov


    """

    if rlambda < 800.0:

        return 1.0

    # convert angstroms to microns

    lambda_mu = rlambda / 1.0e4

    # compute terms of sum

    ratio = lambda_mu / lamb

    term = np.power(ratio, n)

    inv_term = 1.0 / term

    bottom = term + inv_term + b

    xi = np.sum(a / bottom)

    # remove a_b normalization on the extinction curve

    a_lambda = a_b * xi

    # linearize extinction factor

    return math.pow(10.0, -a_lambda / 2.512)


class Standard_Rv(Enum):

    MW = 3.08
    LMC = 3.16
    SMC = 2.93


# create some frozen data classes to hold the extinction curves


_mw_extinction = _ExtinctionCurve(
    a=np.array([165.0, 14.0, 0.045, 0.002, 0.002, 0.012]),
    lamb=np.array([0.047, 0.08, 0.22, 9.7, 18.0, 25.0]),
    b=np.array([90.0, 4.0, -1.95, -1.95, -1.8, 0.0]),
    n=np.array([2.0, 6.5, 2.0, 2.0, 2.0, 2.0]),
)

_lmc_extinction = _ExtinctionCurve(
    a=np.array([175.0, 19.0, 0.023, 0.005, 0.062, 0.02]),
    lamb=np.array([0.046, 0.08, 0.22, 9.7, 18.0, 25.0]),
    b=np.array([90.0, 5.5, -1.95, -1.95, -1.8, 0.0]),
    n=np.array([2.0, 4.5, 2.0, 2.0, 2.0, 2.0]),
)


_smc_extinction = _ExtinctionCurve(
    a=np.array([185.0, 27.0, 0.005, 0.01, 0.012, 0.03]),
    lamb=np.array([0.042, 0.08, 0.22, 9.7, 18.0, 25.0]),
    b=np.array([90.0, 5.5, -1.95, -1.95, -1.8, 0.0]),
    n=np.array([2.0, 4.0, 2.0, 2.0, 2.0, 2.0]),
)


_extinctions_laws: Dict[str, _ExtinctionCurve] = dict(
    mw=_mw_extinction, lmc=_lmc_extinction, smc=_smc_extinction
)



from __future__ import division

import astropy.units as astropy_units
import numpy as np
from past.utils import old_div

import astromodels.functions.numba_functions as nb_func
from astromodels.core.units import get_units
from astromodels.functions.function import (
    Function1D,
    FunctionMeta,
    ModelAssertionViolation,
)


class Blackbody(Function1D, metaclass=FunctionMeta):
    r"""

    description :
        A blackbody function

    latex : $f(x) = K \frac{x^2}{\exp(\frac{x}{kT}) -1}  $

    parameters :
        K :
            desc :
            initial value : 1e-4
            min : 0.
            is_normalization : True

        kT :
            desc : temperature of the blackbody
            initial value : 30.0
            min: 0.

    """

    def _set_units(self, x_unit, y_unit):
        # The normalization has the same units as y
        self.K.unit = old_div(y_unit, (x_unit**2))

        # The break point has always the same dimension as the x variable
        self.kT.unit = x_unit

    def evaluate(self, x, K, kT):

        if isinstance(x, astropy_units.Quantity):

            K_ = K.value
            kT_ = kT.value

            x_ = x.value

            unit_ = self.y_unit

        else:
            unit_ = 1.0
            K_, kT_, x_, = (
                K,
                kT,
                x,
            )

        result = nb_func.bb_eval(x_, K_, kT_)

        return result * unit_


class ModifiedBlackbody(Function1D, metaclass=FunctionMeta):
    r"""

    description :
        A blackbody function

    latex : $f(x) = K \frac{x^2}{\exp(\frac{x}{kT}) -1}  $

    parameters :
        K :
            desc :
            initial value : 1e-4
            min : 0.
            is_normalization : True

        kT :
            desc : temperature of the blackbody
            initial value : 30.0
            min: 0.

    """

    def _set_units(self, x_unit, y_unit):
        # The normalization has the same units as y
        self.K.unit = y_unit / x_unit

        # The break point has always the same dimension as the x variable
        self.kT.unit = x_unit

    def evaluate(self, x, K, kT):

        if isinstance(x, astropy_units.Quantity):

            K_ = K.value
            kT_ = kT.value

            x_ = x.value

            unit_ = self.y_unit

        else:

            unit_ = 1.0
            K_, kT_, x_, = (
                K,
                kT,
                x,
            )

        result = nb_func.mbb_eval(x_, K_, kT_)

        return result * unit_


class NonDissipativePhotosphere(Function1D, metaclass=FunctionMeta):
    r"""

    description :
        Non-dissipative photosphere of a GRB occuring above the saturation radius
        Acuner, Z., Ryde, F. & Yu, H.-F. Mon Not R Astron Soc 487, 5508–5519 (2019).
    latex : $N_{\mathrm{E}}=K\left(\frac{E}{E_{\mathrm{pivot}}}\right)^{0.4} e^{-\left(\frac{E}{E_{c}}\right)^{0.65}}$

    parameters :
        K :
            desc :
            initial value : 1e-4
            min : 0.
            is_normalization : True

        ec :
            desc : peak energy
            initial value : 200.0
            min: 0.

        piv :
            desc : the pivot energy
            initial value: 100.
            fix: True


    """

    def _set_units(self, x_unit, y_unit):
        # The normalization has the same units as y
        self.K.unit = y_unit

        # The break point has always the same dimension as the x variable
        self.ec.unit = x_unit

        self.piv.unit = x_unit

    def evaluate(self, x, K, ec, piv):

        if isinstance(x, astropy_units.Quantity):

            K_ = K.value
            ec_ = ec.value
            piv_ = piv.value

            x_ = x.value

            unit_ = self.y_unit

        else:
            unit_ = 1.0
            K_, ec_, x_, piv_ = (K, ec, x, piv)

        result = nb_func.non_diss_photoshere_generic(x_, K_, ec_, piv_, 0.4, 0.65)

        return result * unit_


class NonDissipativePhotosphere_Deep(Function1D, metaclass=FunctionMeta):
    r"""

    description :
        Non-dissipative photosphere of a GRB occuring BELOW the saturation radius.
        Acuner, Z., Ryde, F. & Yu, H.-F. Mon Not R Astron Soc 487, 5508–5519 (2019).


    latex : $N_{\mathrm{E}}=K\left(\frac{E}{E_{\mathrm{pivot}}}\right)^{0.66} e^{-\left(\frac{E}{E_{c}}\right)}$

    parameters :
        K :
            desc :
            initial value : 1e-4
            min : 0.
            is_normalization : True

        ec :
            desc : peak energy
            initial value : 200.0
            min: 0.

        piv :
            desc : the pivot energy
            initial value: 100.
            fix: True


    """

    def _set_units(self, x_unit, y_unit):
        # The normalization has the same units as y
        self.K.unit = y_unit

        # The break point has always the same dimension as the x variable
        self.ec.unit = x_unit

        self.piv.unit = x_unit

    def evaluate(self, x, K, ec, piv):

        if isinstance(x, astropy_units.Quantity):

            K_ = K.value
            ec_ = ec.value
            piv_ = piv.value

            x_ = x.value

            unit_ = self.y_unit

        else:
            unit_ = 1.0
            K_, ec_, x_, piv_ = (K, ec, x, piv)

        result = nb_func.non_diss_photoshere_generic(x_, K_, ec_, piv_, 0.66, 1.0)

        return result * unit_
